import re
import datetime as dt
from dateutil.relativedelta import relativedelta

import sqlalchemy as sa

from mindsdb_sql import parse_sql, ParsingException

from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.exception import EntityNotExistsError, EntityExistsError
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.query_context.context_controller import query_context_controller
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def split_sql(sql):
    # split sql by ';' ignoring delimiter in quotes
    pattern = re.compile(r'''((?:[^;"']|"[^"]*"|'[^']*')+)''')
    return pattern.split(sql)[1::2]


def calc_next_date(schedule_str, base_date: dt.datetime):
    schedule_str = schedule_str.lower().strip()

    repeat_prefix = 'every '
    if schedule_str.startswith(repeat_prefix):
        repeat_str = schedule_str[len(repeat_prefix):]
    else:
        # TODO cron format
        raise NotImplementedError(f'Schedule: {schedule_str}')

    items = repeat_str.split()

    if len(items) == 1:
        value = '1'
        period = items[0]
    elif len(items) == 2:
        value, period = items
    else:
        raise Exception(f"Can't parse repeat string: {repeat_str}")

    if not value.isdigit():
        raise Exception(f"Number expected: {value}")
    value = int(value)
    if period in ('minute', 'minutes', 'min'):
        delta = dt.timedelta(minutes=value)
    elif period in ('hour', 'hours'):
        delta = dt.timedelta(hours=value)
    elif period in ('day', 'days'):
        delta = dt.timedelta(days=value)
    elif period in ('week', 'weeks'):
        delta = dt.timedelta(days=value * 7)  # 1 week = 7 days
    elif period in ('month', 'months'):
        delta = relativedelta(months=value)
    else:
        raise Exception(f"Unknown period: {period}")

    # period limitation disabled for now
    # config = Config()
    # is_cloud = config.get('cloud', False)
    # if is_cloud and ctx.user_class == 0:
    #     if delta < dt.timedelta(days=1):
    #         raise Exception("Minimal allowed period can't be less than one day")

    next_date = base_date + delta

    return next_date


class JobsController:
    def add(self, name, project_name, query_str, start_at, end_at, repeat_str):

        if project_name is None:
            project_name = 'mindsdb'
        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        # check if exists
        record = db.session.query(db.Jobs).filter_by(
            company_id=ctx.company_id,
            name=name,
            project_id=project.id,
            deleted_at=sa.null()
        ).first()
        if record is not None:
            raise EntityExistsError(f'Job already exists: {name}')

        if start_at is not None:
            start_at = self._parse_date(start_at)
            if start_at < dt.datetime.now():
                start_at = dt.datetime.now()
        else:
            start_at = dt.datetime.now()

        if end_at is not None:
            end_at = self._parse_date(end_at)

            if end_at < start_at:
                raise Exception(f'Wrong end date {start_at} > {end_at}')

        # check sql = try to parse it
        for sql in split_sql(query_str):
            try:
                # replace template variables with null
                sql = re.sub(r'\{\{[\w\d]+}}', "", sql)

                parse_sql(sql, dialect='mindsdb')
            except ParsingException as e:
                raise ParsingException(f'Unable to parse: {sql}: {e}')

        # plan next run
        next_run_at = start_at

        schedule_str = None
        if repeat_str is not None:
            schedule_str = 'every ' + repeat_str

            # try to calculate schedule string
            calc_next_date(schedule_str, start_at)
        else:
            # no schedule for job end_at is meaningless
            end_at = None

        # create job record
        record = db.Jobs(
            company_id=ctx.company_id,
            user_class=ctx.user_class,
            name=name,
            project_id=project.id,
            query_str=query_str,
            start_at=start_at,
            end_at=end_at,
            next_run_at=next_run_at,
            schedule_str=schedule_str
        )
        db.session.add(record)
        db.session.commit()

    def _parse_date(self, date_str):

        if date_str.upper() == 'NOW':
            return dt.datetime.now()

        date_formats = ['%Y-%m-%d %H:%M:%S', '%Y-%m-%d']
        date = None
        for date_format in date_formats:
            try:
                date = dt.datetime.strptime(date_str, date_format)
            except ValueError:
                pass
        if date is None:
            raise ValueError(f"Can't parse date: {date_str}")
        return date

    def delete(self, name, project_name):

        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        # check if exists
        record = db.session.query(db.Jobs).filter_by(
            company_id=ctx.company_id,
            name=name,
            project_id=project.id,
            deleted_at=sa.null()
        ).first()
        if record is None:
            raise EntityNotExistsError('Job does not exist', name)

        self._delete_record(record)
        db.session.commit()

        # delete context
        query_context_controller.drop_query_context('job', record.id)

    def _delete_record(self, record):
        record.deleted_at = dt.datetime.now()

    def get_list(self, project_name=None):

        query = db.session.query(db.Jobs).filter_by(
            company_id=ctx.company_id,
            deleted_at=sa.null()
        )

        project_controller = ProjectController()
        if project_name is not None:
            project = project_controller.get(name=project_name)
            query = query.filter_by(project_id=project.id)

        data = []
        project_names = {
            i.id: i.name
            for i in project_controller.get_list()
        }
        for record in query:
            data.append({
                'id': record.id,
                'name': record.name,
                'project': project_names[record.project_id],
                'start_at': record.start_at,
                'end_at': record.end_at,
                'next_run_at': record.next_run_at,
                'schedule_str': record.schedule_str,
                'query': record.query_str,
                'variables': query_context_controller.get_context_vars('job', record.id)
            })
        return data

    def get_history(self, project_name=None):
        query = db.session.query(db.JobsHistory, db.Jobs).filter_by(
            company_id=ctx.company_id,
        ).outerjoin(db.Jobs, db.Jobs.id == db.JobsHistory.job_id)

        project_controller = ProjectController()
        if project_name is not None:
            project = project_controller.get(name=project_name)
            query = query.filter_by(project_id=project.id)

        data = []
        project_names = {
            i.id: i.name
            for i in project_controller.get_list()
        }
        for record in query:
            data.append({
                'name': record.Jobs.name,
                'project': project_names[record.Jobs.project_id],
                'run_start': record.JobsHistory.start_at,
                'run_end': record.JobsHistory.end_at,
                'error': record.JobsHistory.error,
                'query': record.JobsHistory.query_str,
            })
        return data


class JobsExecutor:

    def get_next_tasks(self):
        # filter next_run < now
        query = db.session.query(db.Jobs).filter(
            db.Jobs.next_run_at < dt.datetime.now(),
            db.Jobs.deleted_at == sa.null(),
            db.Jobs.active == True,  # noqa
        ).order_by(db.Jobs.next_run_at)

        return query.all()

    def update_task_schedule(self, record):
        # calculate next run

        if record.next_run_at > dt.datetime.now():
            # do nothing, it is already planned in future
            return

        if record.schedule_str is None:
            # not need to run it anymore
            self._delete_record(record)
            return

        next_run_at = calc_next_date(record.schedule_str, base_date=record.next_run_at)

        if next_run_at is None:
            # no need to run it
            self._delete_record(record)
        elif record.end_at is not None and next_run_at > record.end_at:
            self._delete_record(record)
        else:
            # plan next run, but not in the past
            if next_run_at < dt.datetime.now():
                next_run_at = dt.datetime.now()
            record.next_run_at = next_run_at

    def _delete_record(self, record):
        record.deleted_at = dt.datetime.now()

    def lock_record(self, record_id):
        # workaround for several concurrent workers on cloud:
        #  create history record before start of task
        record = db.Jobs.query.get(record_id)

        try:

            history_record = db.JobsHistory(
                job_id=record.id,
                start_at=record.next_run_at,
                company_id=record.company_id
            )

            db.session.add(history_record)
            db.session.commit()

            return history_record.id

        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            db.session.rollback()

            # check if it is an old lock
            history_record = db.JobsHistory.query.filter_by(
                job_id=record.id,
                start_at=record.next_run_at,
                company_id=record.company_id
            ).first()
            if history_record.updated_at < dt.datetime.now() - dt.timedelta(seconds=30):
                db.session.delete(history_record)
                db.session.commit()

        return None

    def execute_task_local(self, record_id, history_id=None):

        record = db.Jobs.query.get(record_id)

        # set up environment

        ctx.set_default()
        ctx.company_id = record.company_id
        if record.user_class is not None:
            ctx.user_class = record.user_class

        query_context_controller.set_context('job', record.id)
        if history_id is None:
            history_record = db.JobsHistory(
                job_id=record.id,
                start_at=dt.datetime.now(),
                company_id=record.company_id,
            )
            db.session.add(history_record)
            db.session.flush()
            history_id = history_record.id
            db.session.commit()

        else:
            history_record = db.JobsHistory.query.get(history_id)

        error = ''

        project_controller = ProjectController()
        project = project_controller.get(record.project_id)
        executed_sql = ''
        for sql in split_sql(record.query_str):
            try:
                #  fill template variables
                if '{{PREVIOUS_START_DATETIME}}' in sql:
                    # get previous run date
                    history_prev = db.session.query(db.JobsHistory.start_at)\
                        .filter(db.JobsHistory.job_id == record.id,
                                db.JobsHistory.id != history_id)\
                        .order_by(db.JobsHistory.id.desc())\
                        .first()
                    if history_prev is None:
                        # start date of the job
                        value = record.created_at
                    else:
                        # fix for twitter: created_at filter must be minimum of 10 seconds prior to the current time
                        value = history_prev.start_at - dt.timedelta(seconds=60)
                    value = value.strftime("%Y-%m-%d %H:%M:%S")
                    sql = sql.replace('{{PREVIOUS_START_DATETIME}}', value)

                if '{{START_DATE}}' in sql:
                    value = history_record.start_at.strftime("%Y-%m-%d")
                    sql = sql.replace('{{START_DATE}}', value)
                if '{{START_DATETIME}}' in sql:
                    value = history_record.start_at.strftime("%Y-%m-%d %H:%M:%S")
                    sql = sql.replace('{{START_DATETIME}}', value)
                query = parse_sql(sql, dialect='mindsdb')

                from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
                from mindsdb.api.mysql.mysql_proxy.executor.executor_commands import ExecuteCommands

                sql_session = SessionController()
                sql_session.database = project.name

                command_executor = ExecuteCommands(sql_session, executor=None)

                executed_sql += sql + '; '

                ret = command_executor.execute_command(query)
                if ret.error_code is not None:
                    error = ret.error_message
                    break
            except Exception as e:
                logger.error(e)
                error = str(e)
                break

        try:
            self.update_task_schedule(record)
        except Exception as e:
            db.session.rollback()
            logger.error(f"Error to update schedule: {e}")
            error += f"Error to update schedule: {e}"

            # stop scheduling
            record.next_run_at = None

        history_record = db.JobsHistory.query.get(history_id)

        if error:
            history_record.error = error
        history_record.end_at = dt.datetime.now()
        history_record.query_str = executed_sql

        db.session.commit()



import re
import os
import datetime as dt
import ast
import time
from collections import defaultdict
import pytz
import io
import requests

import pandas as pd
import tweepy

from mindsdb.utilities import log
from mindsdb.utilities.config import Config

from mindsdb_sql.parser import ast

from mindsdb.integrations.libs.api_handler import APIHandler, APITable, FuncParser
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.integrations.utilities.date_utils import parse_utc_date

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)

class TweetsTable(APITable):

    def select(self, query: ast.Select) -> Response:

        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:

            if op == 'or':
                raise NotImplementedError(f'OR is not supported')
            if arg1 == 'created_at':
                date = parse_utc_date(arg2)
                if op == '>':
                    # "tweets/search/recent" doesn't accept dates earlier than 7 days
                    if (dt.datetime.now(dt.timezone.utc) - date).days > 7:
                        # skip this condition
                        continue
                    params['start_time'] = date
                elif op == '<':
                    params['end_time'] = date
                else:
                    raise NotImplementedError

            elif arg1 == 'query':
                if op == '=':
                    params[arg1] = arg2
                else:
                    NotImplementedError(f'Unknown op: {op}')

            elif arg1 == 'id':
                if op == '>':
                    params['since_id'] = arg2
                elif op == '>=':
                    raise NotImplementedError("Please use 'id > value'")
                elif op == '<':
                    params['until_id'] = arg2
                elif op == '<=':
                    raise NotImplementedError("Please use 'id < value'")
                else:
                    NotImplementedError('Search with "id=" is not implemented')

            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params['max_results'] = query.limit.value

        params['expansions'] = ['author_id', 'in_reply_to_user_id']
        params['tweet_fields'] = ['created_at', 'conversation_id', 'referenced_tweets']
        params['user_fields'] = ['name', 'username']

        if 'query' not in params:
            # search not works without query, use 'mindsdb'
            params['query'] = 'mindsdb'

        result = self.handler.call_twitter_api(
            method_name='search_recent_tweets',
            params=params,
            filters=filters
        )

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        return [
            'id',
            'created_at',
            'text',
            'edit_history_tweet_ids',
            'author_id',
            'author_name',
            'author_username',
            'conversation_id',
            'in_reply_to_tweet_id',
            'in_retweeted_to_tweet_id',
            'in_quote_to_tweet_id',
            'in_reply_to_user_id',
            'in_reply_to_user_name',
            'in_reply_to_user_username',
        ]

    def insert(self, query: ast.Insert):
        # https://docs.tweepy.org/en/stable/client.html#tweepy.Client.create_tweet
        columns = [col.name for col in query.columns]

        insert_params = ('consumer_key', 'consumer_secret', 'access_token', 'access_token_secret')
        for p in insert_params:
            if p not in self.handler.connection_args:
                raise Exception(f'To insert data into Twitter, you need to provide the following parameters when connecting it to MindsDB: {insert_params}')  # noqa

        for row in query.values:
            params = dict(zip(columns, row))

            # split long text over 280 symbols
            max_text_len = 280
            text = params['text']

            # Post image if column media_url is provided, only do this on last tweet
            media_ids = None
            if 'media_url' in params:
                media_url = params.pop('media_url')

                # create an in memory file
                resp = requests.get(media_url)
                img = io.BytesIO(resp.content)

                # upload media to twitter
                api_v1 = self.handler.create_connection(api_version=1)
                content_type = resp.headers['Content-Type']
                file_type = content_type.split('/')[-1]
                media = api_v1.media_upload(filename="img.{file_type}".format(file_type=file_type), file=img)

                media_ids = [media.media_id]


            words = re.split('( )', text)

            messages = []

            text2 = ''
            pattern = r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
            for word in words:
                # replace the links in word to string with the length as twitter short url (23)
                word2 = re.sub(pattern, '-' * 23, word)
                if len(text2) + len(word2) > max_text_len - 3 - 7:  # 3 is for ..., 7 is for (10/11)
                    messages.append(text2.strip())

                    text2 = ''
                text2 += word

            # the last message
            if text2.strip() != '':
                messages.append(text2.strip())

            len_messages = len(messages)
            for i, text in enumerate(messages):
                if i < len_messages - 1:
                    text += '...'
                else:
                    text += ' '
                    # publish media with the last message
                    if media_ids is not None:
                        params['media_ids'] = media_ids

                text += f'({i + 1}/{len_messages})'

                params['text'] = text
                ret = self.handler.call_twitter_api('create_tweet', params)
                inserted_id = ret.id[0]
                params['in_reply_to_tweet_id'] = inserted_id


class TwitterHandler(APIHandler):
    """A class for handling connections and interactions with the Twitter API.

    Attributes:
        bearer_token (str): The consumer key for the Twitter app.
        api (tweepy.API): The `tweepy.API` object for interacting with the Twitter API.

    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        args = kwargs.get('connection_data', {})

        self.connection_args = {}
        handler_config = Config().get('twitter_handler', {})
        for k in ['bearer_token', 'consumer_key', 'consumer_secret',
                  'access_token', 'access_token_secret', 'wait_on_rate_limit']:
            if k in args:
                self.connection_args[k] = args[k]
            elif f'TWITTER_{k.upper()}' in os.environ:
                self.connection_args[k] = os.environ[f'TWITTER_{k.upper()}']
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]

        self.api = None
        self.is_connected = False

        tweets = TweetsTable(self)
        self._register_table('tweets', tweets)

    def create_connection(self, api_version=2):
        if api_version == 1:
            auth = tweepy.OAuthHandler(
                self.connection_args['consumer_key'],
                self.connection_args['consumer_secret']
            )
            auth.set_access_token(
                self.connection_args['access_token'],
                self.connection_args['access_token_secret']
            )
            return tweepy.API(auth)

        return tweepy.Client(**self.connection_args)

    def connect(self, api_version=2):
        """Authenticate with the Twitter API using the API keys and secrets stored in the `consumer_key`, `consumer_secret`, `access_token`, and `access_token_secret` attributes."""  # noqa

        if self.is_connected is True:
            return self.api

        self.api = self.create_connection()

        self.is_connected = True
        return self.api

    def check_connection(self) -> StatusResponse:

        response = StatusResponse(False)

        try:
            api = self.connect()

            # call get_user with unknown id.
            #   it raises an error in case if auth is not success and returns not-found otherwise
            #   api.get_me() is not exposed for OAuth 2.0 App-only authorisation
            api.get_user(id=1)
            response.success = True

        except tweepy.Unauthorized as e:
            response.error_message = f'Error connecting to Twitter api: {e}. Check bearer_token'
            logger.error(response.error_message)

        if response.success is True and len(self.connection_args) > 1:
            # not only bearer_token, check read-write mode (OAuth 2.0 Authorization Code with PKCE)
            try:
                api = self.connect()

                api.get_me()

            except tweepy.Unauthorized as e:
                keys = 'consumer_key', 'consumer_secret', 'access_token', 'access_token_secret'
                response.error_message = f'Error connecting to Twitter api: {e}. Check' + ', '.join(keys)
                logger.error(response.error_message)

                response.success = False

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query_string: str = None):
        method_name, params = FuncParser().from_string(query_string)

        df = self.call_twitter_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def _apply_filters(self, data, filters):
        if not filters:
            return data

        data2 = []
        for row in data:
            add = False
            for op, key, value in filters:
                value2 = row.get(key)
                if isinstance(value, int):
                    # twitter returns ids as string
                    value = str(value)

                if op in ('!=', '<>'):
                    if value == value2:
                        break
                elif op in ('==', '='):
                    if value != value2:
                        break
                elif op == 'in':
                    if not isinstance(value, list):
                        value = [value]
                    if value2 not in value:
                        break
                elif op == 'not in':
                    if not isinstance(value, list):
                        value = [value]
                    if value2 in value:
                        break
                else:
                    raise NotImplementedError(f'Unknown filter: {op}')
                # only if there wasn't breaks
                add = True
            if add:
                data2.append(row)
        return data2

    def call_twitter_api(self, method_name: str = None, params: dict = None, filters: list = None):

        # method > table > columns
        expansions_map = {
            'search_recent_tweets': {
                'users': ['author_id', 'in_reply_to_user_id'],
            },
            'search_all_tweets': {
                'users': ['author_id'],
            },
        }

        api = self.connect()
        method = getattr(api, method_name)

        # pagination handle

        count_results = None
        if 'max_results' in params:
            count_results = params['max_results']

        data = []
        includes = defaultdict(list)

        max_page_size = 100
        min_page_size = 10
        left = None

        limit_exec_time = time.time() + 60

        if filters:
            # if we have filters: do big page requests
            params['max_results'] = max_page_size

        while True:
            if time.time() > limit_exec_time:
                raise RuntimeError('Handler request timeout error')

            if count_results is not None:
                left = count_results - len(data)
                if left == 0:
                    break
                elif left < 0:
                    # got more results that we need
                    data = data[:left]
                    break

                if left > max_page_size:
                    params['max_results'] = max_page_size
                elif left < min_page_size:
                    params['max_results'] = min_page_size
                else:
                    params['max_results'] = left

            logger.debug(f'>>>twitter in: {method_name}({params})')
            resp = method(**params)

            if hasattr(resp, 'includes'):
                for table, records in resp.includes.items():
                    includes[table].extend([r.data for r in records])

            if isinstance(resp.data, list):
                chunk = [r.data for r in resp.data]
            else:
                if isinstance(resp.data, dict):
                    data.append(resp.data)
                if hasattr(resp.data, 'data') and isinstance(resp.data.data, dict):
                    data.append(resp.data.data)
                break

            # unwind columns
            for row in chunk:
                if 'referenced_tweets' in row:
                    refs = row['referenced_tweets']
                    if isinstance(refs, list) and len(refs) > 0:
                        if refs[0]['type'] == 'replied_to':
                            row['in_reply_to_tweet_id'] = refs[0]['id']
                        if refs[0]['type'] == 'retweeted':
                            row['in_retweeted_to_tweet_id'] = refs[0]['id']
                        if refs[0]['type'] == 'quoted':
                            row['in_quote_to_tweet_id'] = refs[0]['id']

            if filters:
                chunk = self._apply_filters(chunk, filters)

            # limit output
            if left is not None:
                chunk = chunk[:left]

            data.extend(chunk)
            # next page ?
            if count_results is not None and hasattr(resp, 'meta') and 'next_token' in resp.meta:
                params['next_token'] = resp.meta['next_token']
            else:
                break

        df = pd.DataFrame(data)

        # enrich
        expansions = expansions_map.get(method_name)
        if expansions is not None:
            for table, records in includes.items():
                df_ref = pd.DataFrame(records)

                if table not in expansions:
                    continue

                for col_id in expansions[table]:
                    col = col_id[:-3]  # cut _id
                    if col_id not in df.columns:
                        continue

                    col_map = {
                        col_ref: f'{col}_{col_ref}'
                        for col_ref in df_ref.columns
                    }
                    df_ref2 = df_ref.rename(columns=col_map)
                    df_ref2 = df_ref2.drop_duplicates(col_id)

                    df = df.merge(df_ref2, on=col_id, how='left')

        return df



from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql.parser import ast
from mindsdb.integrations.utilities.date_utils import parse_local_date
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions, project_dataframe, filter_dataframe
from mindsdb.integrations.utilities.sql_utils import sort_dataframe
from mindsdb.integrations.utilities.utils import dict_to_yaml
from typing import Dict, List

import pandas as pd


class OpenBBtable(APITable):
    def _get_params_from_conditions(self, conditions: List) -> Dict:
        """Gets aggregate trade data API params from SQL WHERE conditions.

        Returns params to use for Binance API call to klines.

        Args:
            conditions (List): List of individual SQL WHERE conditions.
        """
        params: dict = {}
        # generic interpreter for conditions
        # since these are all equality conditions due to OpenBB Platform's API
        # then we can just use the first arg as the key and the second as the value
        for op, arg1, arg2 in conditions:
            if op != "=":
                raise NotImplementedError
            params[arg1] = arg2

        return params

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Selects data from the OpenBB Platform and returns it as a pandas DataFrame.

        Returns dataframe representing the OpenBB data.

        Args:
            query (ast.Select): Given SQL SELECT query
        """
        conditions = extract_comparison_conditions(query.where)
        params = self._get_params_from_conditions(conditions)

        openbb_data = self.handler.call_openbb_api(
            method_name="openbb_fetcher",
            params=params,
        )

        return openbb_data


def create_table_class(params_metadata, response_metadata, obb_function):

    mandatory_fields = params_metadata['required'] if 'required' in params_metadata else []
    response_columns = list(response_metadata['properties'].keys())

    class AnyTable(APITable):
        def _get_params_from_conditions(self, conditions: List) -> Dict:
            """Gets aggregate trade data API params from SQL WHERE conditions.

            Returns params to use for Binance API call to klines.

            Args:
                conditions (List): List of individual SQL WHERE conditions.
            """
            params: dict = {}
            # generic interpreter for conditions
            # since these are all equality conditions due to OpenBB Platform's API
            # then we can just use the first arg as the key and the second as the value
            for op, arg1, arg2 in conditions:
                if op == "=":
                    params[arg1] = arg2

            return params

        def select(self, query: ast.Select) -> pd.DataFrame:
            """Selects data from the OpenBB Platform and returns it as a pandas DataFrame.

            Returns dataframe representing the OpenBB data.

            Args:
                query (ast.Select): Given SQL SELECT query
            """
            conditions = extract_comparison_conditions(query.where)
            arg_params = self._get_params_from_conditions(conditions=conditions)
            params = {}
            filters = []
            mandatory_args = {key: False for key in mandatory_fields}
            columns_to_add = {}
            # filter only for properties in the
            strict_filter = arg_params.get('strict_filter', False)
            for op, arg1, arg2 in conditions:

                if op == 'or':
                    raise NotImplementedError('OR is not supported')
                if arg1 in mandatory_fields:
                    mandatory_args[arg1] = True

                if ('start_' + arg1 in params_metadata['properties']
                    and arg1 in response_columns and arg2 is not None
                        and "format" in response_metadata['properties'][arg1]):
                    if response_metadata['properties'][arg1]["format"] != 'date-time':
                        date = parse_local_date(arg2)
                        interval = arg_params.get('interval', '1d')
                        if op == '>':
                            params['start_' + arg1] = date.strftime('%Y-%m-%d')
                        elif op == '<':
                            params['end_' + arg1] = date.strftime('%Y-%m-%d')
                        elif op == '>=':
                            date = date - pd.Timedelta(interval)
                            params['start_' + arg1] = date.strftime('%Y-%m-%d')
                        elif op == '<=':
                            date = date + pd.Timedelta(interval)
                            params['end_' + arg1] = date.strftime('%Y-%m-%d')
                        elif op == '=':
                            date = date - pd.Timedelta(interval)
                            params['start_' + arg1] = date.strftime('%Y-%m-%d')
                            date = date + pd.Timedelta(interval)
                            params['end_' + arg1] = date.strftime('%Y-%m-%d')
                elif arg1 in params_metadata['properties'] or not strict_filter:
                    if op == '=':
                        params[arg1] = arg2
                        columns_to_add[arg1] = arg2
                filters.append([op, arg1, arg2])

            if not all(mandatory_args.values()):
                string = 'You must specify the following arguments in the WHERE statement:'
                for key in mandatory_args:
                    if not mandatory_args[key]:
                        string += "\n--(required)---\n* {key}:\n{help}\n ".format(key=key, help=dict_to_yaml(params_metadata['properties'][key]))
                for key in params_metadata["properties"]:
                    if key not in mandatory_args:
                        string += "\n--(optional)---\n* {key}:\n{help}\n ".format(key=key, help=dict_to_yaml(params_metadata['properties'][key]))
                raise NotImplementedError(string)

            result = obb_function(**params).to_df()

            # Check if index is a datetime, if it is we want that as a column
            if isinstance(result.index, pd.DatetimeIndex):
                result.reset_index(inplace=True)

            if query.limit is not None:
                result = result.head(query.limit.value)

            for key in columns_to_add:
                result[key] = params[key]
            # filter targets
            result = filter_dataframe(result, filters)

            columns = self.get_columns()

            columns += [col for col in result.columns if col not in columns]

            # project targets
            result = project_dataframe(result, query.targets, columns)
            # test this
            if query.order_by:
                result = sort_dataframe(result, query.order_by)

            return result

        def get_columns(self):
            return response_columns
    return AnyTable



import os
from typing import Optional, Dict

import openai
import pandas as pd
from langchain.llms import OpenAI
import llama_index
from llama_index.readers.schema.base import Document
from llama_index import SimpleWebPageReader, QuestionAnswerPrompt
from llama_index import ServiceContext, StorageContext, load_index_from_storage
from llama_index import LLMPredictor, OpenAIEmbedding
from llama_index.indices.vector_store.base import VectorStore

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities.config import Config
from mindsdb.utilities.security import is_private_url


def _validate_prompt_template(prompt_template: str):
    if '{context_str}' not in prompt_template or '{query_str}' not in prompt_template:
        raise Exception(
            "Provided prompt template is invalid, missing `{context_str}`, `{query_str}`. Please ensure both placeholders are present and try again.")  # noqa


class LlamaIndexHandler(BaseMLEngine):
    """ Integration with the LlamaIndex data framework for LLM applications. """
    name = 'llama_index'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.default_index_class = 'GPTVectorStoreIndex'
        self.supported_index_class = ['GPTVectorStoreIndex']
        self.default_reader = 'DFReader'
        self.supported_reader = ['DFReader', 'SimpleWebPageReader']

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'prompt_template' in args['using']:
            _validate_prompt_template(args['using']['prompt_template'])

        if args['using'].get('mode') == 'conversational':
            for param in ('user_column', 'assistant_column'):
                if param not in args['using']:
                    raise Exception(f'Conversational mode requires {param} parameter')

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if 'using' not in args:
            raise Exception("LlamaIndex engine requires a USING clause! Refer to its documentation for more details.")

        if 'index_class' not in args['using']:
            args['using']['index_class'] = self.default_index_class
        elif args['using']['index_class'] not in self.supported_index_class:
            raise Exception(f"Invalid index class argument. Please use one of {self.supported_index_class}")

        if 'reader' not in args['using']:
            args['using']['reader'] = self.default_reader
        elif args['using']['reader'] not in self.supported_reader:
            raise Exception(f"Invalid operation mode. Please use one of {self.supported_reader}")

        # workaround to create llama model without input data
        if df is None or df.empty:
            df = pd.DataFrame([{'text': ''}])

        if args['using']['reader'] == 'DFReader':
            dstrs = df.apply(lambda x: ', '.join([f'{col}: {str(entry)}' for col, entry in zip(df.columns, x)]), axis=1)
            reader = list(map(lambda x: Document(text=x), dstrs.tolist()))

        elif args['using']['reader'] == 'SimpleWebPageReader':
            if 'source_url_link' not in args['using']:
                raise Exception("SimpleWebPageReader requires a `source_url_link` parameter. Refer to LlamaIndex documentation for more details.")  # noqa

            url = args['using']['source_url_link']
            config = Config()
            is_cloud = config.get("cloud", False)
            if is_cloud and is_private_url(url):
                raise Exception(f'URL is private: {url}')

            reader = SimpleWebPageReader(html_to_text=True).load_data([url])

        else:
            raise Exception(f"Invalid operation mode. Please use one of {self.supported_reader}.")

        self.model_storage.json_set('args', args)
        index = self._setup_index(reader)
        path = self.model_storage.folder_get('context')
        index.storage_context.persist(persist_dir=path)
        self.model_storage.folder_sync('context')

    def update(self, args) -> None:
        prompt_template = args['using'].get('prompt_template', args.get('prompt_template', None))
        if prompt_template is not None:
            _validate_prompt_template(prompt_template)
        args_cur = self.model_storage.json_get('args')
        args_cur['using'].update(args['using'])

        # check new set of arguments
        self.create_validation(None, args_cur)

        self.model_storage.json_set('args', args_cur)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> pd.DataFrame:
        pred_args = args['predict_params'] if args else {}

        args = self.model_storage.json_get('args')
        engine_kwargs = {}

        if args['using'].get('mode') == 'conversational':
            user_column = args['using']['user_column']
            assistant_column = args['using']['assistant_column']

            messages = []
            for row in df[:-1].to_dict('records'):

                messages.append(f'user: {row[user_column]}')
                messages.append(f'assistant: {row[assistant_column]}')

            conversation = '\n'.join(messages)

            questions = [
                df.iloc[-1][user_column]
            ]

            if 'prompt' in pred_args and pred_args['prompt'] is not None:
                user_prompt = pred_args['prompt']
            else:
                user_prompt = args['using'].get('prompt', '')

            prompt_template = f'{user_prompt}\n'\
                f'---------------------\n' \
                f'We have provided context information below. \n' \
                f'{{context_str}}\n' \
                f'---------------------\n' \
                f'This is previous conversation history:\n' \
                f'{conversation}\n' \
                f'---------------------\n' \
                f'Given this information, please answer the question: {{query_str}}'

            engine_kwargs['text_qa_template'] = QuestionAnswerPrompt(prompt_template)

        else:
            input_column = args['using'].get('input_column', None)

            prompt_template = args['using'].get('prompt_template', args.get('prompt_template', None))
            if prompt_template is not None:
                _validate_prompt_template(prompt_template)
                engine_kwargs['text_qa_template'] = QuestionAnswerPrompt(prompt_template)

            if input_column is None:
                raise Exception(f'`input_column` must be provided at model creation time or through USING clause when predicting. Please try again.')  # noqa

            if input_column not in df.columns:
                raise Exception(f'Column "{input_column}" not found in input data! Please try again.')

            questions = df[input_column]

        index_path = self.model_storage.folder_get('context')
        storage_context = StorageContext.from_defaults(persist_dir=index_path)
        service_context = self._get_service_context()
        index = load_index_from_storage(storage_context, service_context=service_context)
        query_engine = index.as_query_engine(**engine_kwargs)

        results = []

        for question in questions:
            query_results = query_engine.query(question)  # TODO: provide extra_info in explain_target col
            results.append(query_results.response)

        result_df = pd.DataFrame({'question': questions, args['target']: results})  # result_df['answer'].tolist()
        return result_df

    def _get_service_context(self):
        args = self.model_storage.json_get('args')
        openai_api_key = self._get_llama_index_api_key(args['using'])
        openai.api_key = openai_api_key  # TODO: shouldn't have to do this! bug?
        llm_kwargs = {
            'openai_api_key': openai_api_key
        }
        if 'temperature' in args['using']:
            llm_kwargs['temperature'] = args['using']['temperature']
        if 'model_name' in args['using']:
            llm_kwargs['model_name'] = args['using']['model_name']
        if 'max_tokens' in args['using']:
            llm_kwargs['max_tokens'] = args['using']['max_tokens']

        llm = OpenAI(**llm_kwargs)  # TODO: all usual params should go here
        embed_model = OpenAIEmbedding(openai_api_key=openai_api_key)
        service_context = ServiceContext.from_defaults(
            llm_predictor=LLMPredictor(llm=llm),
            embed_model=embed_model
        )
        return service_context
    
    def _setup_index(self, documents):
        args = self.model_storage.json_get('args')
        indexer: VectorStore = getattr(llama_index, args['using']['index_class'])
        index = indexer.from_documents(documents, service_context=self._get_service_context())

        return index

    def _get_llama_index_api_key(self, args, strict=True):
        """
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. OPENAI_API_KEY env variable
            4. llama_index.OPENAI_API_KEY setting in config.json

        Note: method is not case sensitive.
        """
        key = 'OPENAI_API_KEY'
        for k in key, key.lower():
            # 1
            if args.get(k):
                return args[k]
            # 2
            connection_args = self.engine_storage.get_connection_args()
            if k in connection_args:
                return connection_args[k]
            # 3
            api_key = os.getenv(k)
            if api_key is not None:
                return api_key
            # 4
            config = Config()
            openai_cfg = config.get('llama_index', {})
            if k in openai_cfg:
                return openai_cfg[k]

        if strict:
            raise Exception(f'Missing API key "{k}". Either re-create this ML_ENGINE specifying the `{k}` parameter, or re-create this model and pass the API key with `USING` syntax.')  # noqa



PDF_CONTENT = (
    b"%PDF-1.7\n\n1 0 obj\n<< /Type /Catalog /Pages 2 0 R >>\nendobj\n\n2 0 obj\n<< /Type /Pages "
    b"/Kids [3 0 R] /Count 1 >>\nendobj\n\n3 0 obj\n<< /Type /Page /Parent 2 0 R /Contents 4 0 R "
    b">>\nendobj\n\n4 0 obj\n<< /Length 22 >>\nstream\nBT\n/Helvetica 12 Tf\n1 0 0 1 50 700 Tm\n("
    b"Hello, this is a test!) Tj\nET\nendstream\nendobj\n\nxref\n0 5\n0000000000 65535 "
    b"f\n0000000010 00000 n\n0000000077 00000 n\n0000000122 00000 n\n0000000203 00000 n\n0000000277 "
    b"00000 n\ntrailer\n<< /Size 5 /Root 1 0 R >>\nstartxref\n343\n%%EOF\n "
)

BROKEN_PDF_CONTENT = b"%PDF-1.4\n\nThis is not a valid PDF file content\n"

HTML_SAMPLE_1 = "<h1>Heading One</h1><h2>Heading Two</h2><ul><li>item1</li><li>item2</li><li>item3</li></ul>"

MARKDOWN_SAMPLE_1 = "# Heading One\n\n## Heading Two\n\n* item1\n* item2\n* item3\n\n"

HTML_SAMPLE_2 = '<h3>Heading</h3><p>text</p><a href="https://google.com">link</a><ul><ol>item1</ol></ul>'

MARKDOWN_SAMPLE_2 = "### Heading\n\ntext\n\n[link](https://google.com)\n\n\n\n"



from collections import OrderedDict
from mindsdb_sql import parse_sql

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.handlers.sap_erp_handler.api import SAPERP
from mindsdb.integrations.handlers.sap_erp_handler.sap_erp_tables import (
    AddressEmailAddressTable,
    AddressFaxNumberTable,
    AddressHomePageURLTable,
    AddressPhoneNumberTable,
    BPAddrDepdntIntlLocNumberTable,
    BPContactToAddressTable,
    BPContactToFuncAndDeptTable,
    BPCreditWorthinessTable,
    BPDataControllerTable,
    BPFinancialServicesExtnTable,
    BPFinancialServicesReportingTable,
    BPFiscalYearInformationTable,
    BPRelationshipTable,
    BuPaAddressUsageTable,
    BuPaIdentificationTable,
    BuPaIndustryTable,
    BusinessPartnerTable,
    BusinessPartnerAddressTable,
    BusinessPartnerContactTable,
    BusinessPartnerPaymentCardTable,
    BusinessPartnerRatingTable,
    BusinessPartnerRoleTable,
    BusinessPartnerTaxNumberTable,
    BusPartAddrDepdntTaxNumberTable,
    CustAddrDepdntExtIdentifierTable,
    CustAddrDepdntInformationTable,
    CustomerCompanyTable,
    CustomerCompanyTextTable,
    CustomerDunningTable,
    CustomerSalesAreaTable,
    CustomerSalesAreaTaxTable,
    CustomerSalesAreaTextTable,
    CustomerTaxGroupingTable,
    CustomerTextTable,
    CustomerUnloadingPointTable,
    CustomerWithHoldingTaxTable,
    CustSalesPartnerFuncTable,
    CustSlsAreaAddrDepdntInfoTable,
    CustSlsAreaAddrDepdntTaxInfoTable,
    CustUnldgPtAddrDepdntInfoTable,
    SupplierTable,
    SupplierCompanyTable,
    SupplierCompanyTextTable,
    SupplierDunningTable,
    SupplierPartnerFuncTable,
    SupplierPurchasingOrgTable,
    SupplierPurchasingOrgTextTable,
    SupplierTextTable,
    SupplierWithHoldingTaxTable,
)

from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse


class SAPERPHandler(APIHandler):

    def __init__(self, name: str, **kwargs) -> None:
        super().__init__(name)
        self.connection = None
        self.is_connected = False
        self.api_key = kwargs.get("connection_data", {}).get("api_key", "")
        self.base_url = kwargs.get("connection_data", {}).get("base_url", "")
        _tables = [
            AddressEmailAddressTable,
            AddressFaxNumberTable,
            AddressHomePageURLTable,
            AddressPhoneNumberTable,
            BPAddrDepdntIntlLocNumberTable,
            BPContactToAddressTable,
            BPContactToFuncAndDeptTable,
            BPCreditWorthinessTable,
            BPDataControllerTable,
            BPFinancialServicesExtnTable,
            BPFinancialServicesReportingTable,
            BPFiscalYearInformationTable,
            BPRelationshipTable,
            BuPaAddressUsageTable,
            BuPaIdentificationTable,
            BuPaIndustryTable,
            BusinessPartnerTable,
            BusinessPartnerAddressTable,
            BusinessPartnerContactTable,
            BusinessPartnerPaymentCardTable,
            BusinessPartnerRatingTable,
            BusinessPartnerRoleTable,
            BusinessPartnerTaxNumberTable,
            BusPartAddrDepdntTaxNumberTable,
            CustAddrDepdntExtIdentifierTable,
            CustAddrDepdntInformationTable,
            CustomerCompanyTable,
            CustomerCompanyTextTable,
            CustomerDunningTable,
            CustomerSalesAreaTable,
            CustomerSalesAreaTaxTable,
            CustomerSalesAreaTextTable,
            CustomerTaxGroupingTable,
            CustomerTextTable,
            CustomerUnloadingPointTable,
            CustomerWithHoldingTaxTable,
            CustSalesPartnerFuncTable,
            CustSlsAreaAddrDepdntInfoTable,
            CustSlsAreaAddrDepdntTaxInfoTable,
            CustUnldgPtAddrDepdntInfoTable,
            SupplierTable,
            SupplierCompanyTable,
            SupplierCompanyTextTable,
            SupplierDunningTable,
            SupplierPartnerFuncTable,
            SupplierPurchasingOrgTable,
            SupplierPurchasingOrgTextTable,
            SupplierTextTable,
            SupplierWithHoldingTaxTable,
        ]
        for Table in _tables:
            self._register_table(Table.name, Table(self))
        self.connect()

    def check_connection(self) -> StatusResponse:
        resp = StatusResponse(False)
        if self.connection and not self.connection.is_connected():
            resp.error = "Client not connected"
        else:
            resp.success = True
        return resp

    def connect(self) -> SAPERP:
        self.connection = SAPERP(self.base_url, self.api_key)
        return self.connection

    def native_query(self, query: str) -> StatusResponse:
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    api_key={
        'type': ARG_TYPE.STR,
        'description': 'API Token for accessing SAP ERP',
        'required': True,
        'label': 'API Key',
    },
    base_url={
        'type': ARG_TYPE.STR,
        'description': 'Base URL of the host',
        'required': True,
        'label': 'Base URL',
    }
)


connection_args_example = OrderedDict(
    api_key='23d6b9e0c2fab7eba2e8b7e452cead47',
    base_url='https://sandbox.api.sap.com/s4hanacloud/sap/opu/odata/sap/API_BUSINESS_PARTNER/'
)



from functools import partial

import pandas as pd
from mindsdb_sql.parser.ast import BinaryOperation, Constant, Identifier, Select
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.api.mysql.mysql_proxy.classes.sql_query import get_all_tables
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import (
    TABLES_ROW_TYPE,
    TablesRow,
)
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.integration_datanode import (
    IntegrationDataNode,
)
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.project_datanode import (
    ProjectDataNode,
)
from mindsdb.api.mysql.mysql_proxy.utilities import exceptions as exc
from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.interfaces.agents.agents_controller import AgentsController
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.jobs.jobs_controller import JobsController
from mindsdb.interfaces.skills.skills_controller import SkillsController
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class InformationSchemaDataNode(DataNode):
    type = "INFORMATION_SCHEMA"

    information_schema = {
        "SCHEMATA": [
            "CATALOG_NAME",
            "SCHEMA_NAME",
            "DEFAULT_CHARACTER_SET_NAME",
            "DEFAULT_COLLATION_NAME",
            "SQL_PATH",
        ],
        "TABLES": [
            "TABLE_CATALOG",
            "TABLE_SCHEMA",
            "TABLE_NAME",
            "TABLE_TYPE",
            "ENGINE",
            "VERSION",
            "ROW_FORMAT",
            "TABLE_ROWS",
            "AVG_ROW_LENGTH",
            "DATA_LENGTH",
            "MAX_DATA_LENGTH",
            "INDEX_LENGTH",
            "DATA_FREE",
            "AUTO_INCREMENT",
            "CREATE_TIME",
            "UPDATE_TIME",
            "CHECK_TIME",
            "TABLE_COLLATION",
            "CHECKSUM",
            "CREATE_OPTIONS",
            "TABLE_COMMENT",
        ],
        "COLUMNS": [
            "TABLE_CATALOG",
            "TABLE_SCHEMA",
            "TABLE_NAME",
            "COLUMN_NAME",
            "ORDINAL_POSITION",
            "COLUMN_DEFAULT",
            "IS_NULLABLE",
            "DATA_TYPE",
            "CHARACTER_MAXIMUM_LENGTH",
            "CHARACTER_OCTET_LENGTH",
            "NUMERIC_PRECISION",
            "NUMERIC_SCALE",
            "DATETIME_PRECISION",
            "CHARACTER_SET_NAME",
            "COLLATION_NAME",
            "COLUMN_TYPE",
            "COLUMN_KEY",
            "EXTRA",
            "PRIVILEGES",
            "COLUMN_COMMENT",
            "GENERATION_EXPRESSION",
        ],
        "EVENTS": [
            "EVENT_CATALOG",
            "EVENT_SCHEMA",
            "EVENT_NAME",
            "DEFINER",
            "TIME_ZONE",
            "EVENT_BODY",
            "EVENT_DEFINITION",
            "EVENT_TYPE",
            "EXECUTE_AT",
            "INTERVAL_VALUE",
            "INTERVAL_FIELD",
            "SQL_MODE",
            "STARTS",
            "ENDS",
            "STATUS",
            "ON_COMPLETION",
            "CREATED",
            "LAST_ALTERED",
            "LAST_EXECUTED",
            "EVENT_COMMENT",
            "ORIGINATOR",
            "CHARACTER_SET_CLIENT",
            "COLLATION_CONNECTION",
            "DATABASE_COLLATION",
        ],
        "ROUTINES": [
            "SPECIFIC_NAME",
            "ROUTINE_CATALOG",
            "ROUTINE_SCHEMA",
            "ROUTINE_NAME",
            "ROUTINE_TYPE",
            "DATA_TYPE",
            "CHARACTER_MAXIMUM_LENGTH",
            "CHARACTER_OCTET_LENGTH",
            "NUMERIC_PRECISION",
            "NUMERIC_SCALE",
            "DATETIME_PRECISION",
            "CHARACTER_SET_NAME",
            "COLLATION_NAME",
            "DTD_IDENTIFIER",
            "ROUTINE_BODY",
            "ROUTINE_DEFINITION",
            "EXTERNAL_NAME",
            "EXTERNAL_LANGUAGE",
            "PARAMETER_STYLE",
            "IS_DETERMINISTIC",
            "SQL_DATA_ACCESS",
            "SQL_PATH",
            "SECURITY_TYPE",
            "CREATED",
            "LAST_ALTERED",
            "SQL_MODE",
            "ROUTINE_COMMENT",
            "DEFINER",
            "CHARACTER_SET_CLIENT",
            "COLLATION_CONNECTION",
            "DATABASE_COLLATION",
        ],
        "TRIGGERS": [
            "TRIGGER_CATALOG",
            "TRIGGER_SCHEMA",
            "TRIGGER_NAME",
            "EVENT_MANIPULATION",
            "EVENT_OBJECT_CATALOG",
            "EVENT_OBJECT_SCHEMA",
            "EVENT_OBJECT_TABLE",
            "ACTION_ORDER",
            "ACTION_CONDITION",
            "ACTION_STATEMENT",
            "ACTION_ORIENTATION",
            "ACTION_TIMING",
            "ACTION_REFERENCE_OLD_TABLE",
            "ACTION_REFERENCE_NEW_TABLE",
            "ACTION_REFERENCE_OLD_ROW",
            "ACTION_REFERENCE_NEW_ROW",
            "CREATED",
            "SQL_MODE",
            "DEFINER",
            "CHARACTER_SET_CLIENT",
            "COLLATION_CONNECTION",
            "DATABASE_COLLATION",
        ],
        "PLUGINS": [
            "PLUGIN_NAME",
            "PLUGIN_VERSION",
            "PLUGIN_STATUS",
            "PLUGIN_TYPE",
            "PLUGIN_TYPE_VERSION",
            "PLUGIN_LIBRARY",
            "PLUGIN_LIBRARY_VERSION",
            "PLUGIN_AUTHOR",
            "PLUGIN_DESCRIPTION",
            "PLUGIN_LICENSE",
            "LOAD_OPTION",
            "PLUGIN_MATURITY",
            "PLUGIN_AUTH_VERSION",
        ],
        "ENGINES": ["ENGINE", "SUPPORT", "COMMENT", "TRANSACTIONS", "XA", "SAVEPOINTS"],
        "KEY_COLUMN_USAGE": [
            "CONSTRAINT_CATALOG",
            "CONSTRAINT_SCHEMA",
            "CONSTRAINT_NAME",
            "TABLE_CATALOG",
            "TABLE_SCHEMA",
            "TABLE_NAME",
            "COLUMN_NAME",
            "ORDINAL_POSITION",
            "POSITION_IN_UNIQUE_CONSTRAINT",
            "REFERENCED_TABLE_SCHEMA",
            "REFERENCED_TABLE_NAME",
            "REFERENCED_COLUMN_NAME",
        ],
        "STATISTICS": [
            "TABLE_CATALOG",
            "TABLE_SCHEMA",
            "TABLE_NAME",
            "NON_UNIQUE",
            "INDEX_SCHEMA",
            "INDEX_NAME",
            "SEQ_IN_INDEX",
            "COLUMN_NAME",
            "COLLATION",
            "CARDINALITY",
            "SUB_PART",
            "PACKED",
            "NULLABLE",
            "INDEX_TYPE",
            "COMMENT",
            "INDEX_COMMENT",
            "IS_VISIBLE",
            "EXPRESSION",
        ],
        "CHARACTER_SETS": [
            "CHARACTER_SET_NAME",
            "DEFAULT_COLLATE_NAME",
            "DESCRIPTION",
            "MAXLEN",
        ],
        "COLLATIONS": [
            "COLLATION_NAME",
            "CHARACTER_SET_NAME",
            "ID",
            "IS_DEFAULT",
            "IS_COMPILED",
            "SORTLEN",
            "PAD_ATTRIBUTE",
        ],
        # MindsDB specific:
        "MODELS": [
            "NAME",
            "ENGINE",
            "PROJECT",
            "VERSION",
            "STATUS",
            "ACCURACY",
            "PREDICT",
            "UPDATE_STATUS",
            "MINDSDB_VERSION",
            "ERROR",
            "SELECT_DATA_QUERY",
            "TRAINING_OPTIONS",
            "CURRENT_TRAINING_PHASE",
            "TOTAL_TRAINING_PHASES",
            "TRAINING_PHASE_NAME",
            "TAG",
            "CREATED_AT",
            "TRAINING_TIME",
        ],
        "MODELS_VERSIONS": [
            "NAME",
            "ENGINE",
            "PROJECT",
            "ACTIVE",
            "VERSION",
            "STATUS",
            "ACCURACY",
            "PREDICT",
            "UPDATE_STATUS",
            "MINDSDB_VERSION",
            "ERROR",
            "SELECT_DATA_QUERY",
            "TRAINING_OPTIONS",
            "TAG",
            "CREATED_AT",
            "TRAINING_TIME",
        ],
        "DATABASES": ["NAME", "TYPE", "ENGINE", "CONNECTION_DATA"],
        "ML_ENGINES": ["NAME", "HANDLER", "CONNECTION_DATA"],
        "HANDLERS": [
            "NAME",
            "TYPE",
            "TITLE",
            "DESCRIPTION",
            "VERSION",
            "CONNECTION_ARGS",
            "IMPORT_SUCCESS",
            "IMPORT_ERROR",
        ],
        "JOBS": [
            "NAME",
            "PROJECT",
            "START_AT",
            "END_AT",
            "NEXT_RUN_AT",
            "SCHEDULE_STR",
            "QUERY",
            "VARIABLES",
        ],
        "MDB_TRIGGERS": ["NAME", "PROJECT", "DATABASE", "TABLE", "QUERY", "LAST_ERROR"],
        "JOBS_HISTORY": ["NAME", "PROJECT", "RUN_START", "RUN_END", "ERROR", "QUERY"],
        "CHATBOTS": [
            "NAME",
            "PROJECT",
            "DATABASE",
            "MODEL_NAME",
            "PARAMS",
            "IS_RUNNING",
            "LAST_ERROR",
        ],
        "KNOWLEDGE_BASES": ["NAME", "PROJECT", "MODEL", "STORAGE"],
        "SKILLS": ["NAME", "PROJECT", "TYPE", "PARAMS"],
        "AGENTS": [
            "NAME",
            "PROJECT",
            "MODEL_NAME",
            "SKILLS",
            "PARAMS"
        ]
    }

    def __init__(self, session):
        self.session = session
        self.integration_controller = session.integration_controller
        self.project_controller = ProjectController()
        self.database_controller = session.database_controller

        self.persis_datanodes = {}

        databases = self.database_controller.get_dict()
        if "files" in databases:
            self.persis_datanodes["files"] = IntegrationDataNode(
                "files",
                ds_type="file",
                integration_controller=self.session.integration_controller,
            )

        self.get_dataframe_funcs = {
            "TABLES": self._get_tables,
            "COLUMNS": self._get_columns,
            "SCHEMATA": self._get_schemata,
            "ENGINES": self._get_engines,
            "CHARACTER_SETS": self._get_charsets,
            "COLLATIONS": self._get_collations,
            "MODELS": self._get_models,
            "MODELS_VERSIONS": self._get_models_versions,
            "DATABASES": self._get_databases,
            "ML_ENGINES": self._get_ml_engines,
            "HANDLERS": self._get_handlers,
            "JOBS": self._get_jobs,
            "JOBS_HISTORY": self._get_jobs_history,
            "MDB_TRIGGERS": self._get_triggers,
            "CHATBOTS": self._get_chatbots,
            "KNOWLEDGE_BASES": self._get_knowledge_bases,
            "SKILLS": self._get_skills,
            "AGENTS": self._get_agents
        }
        for table_name in self.information_schema:
            if table_name not in self.get_dataframe_funcs:
                self.get_dataframe_funcs[table_name] = partial(
                    self._get_empty_table, table_name
                )

    def __getitem__(self, key):
        return self.get(key)

    def get(self, name):
        name_lower = name.lower()

        if name.lower() == "information_schema":
            return self

        if name_lower in self.persis_datanodes:
            return self.persis_datanodes[name_lower]

        existing_databases_meta = (
            self.database_controller.get_dict()
        )  # filter_type='project'
        database_name = None
        for key in existing_databases_meta:
            if key.lower() == name_lower:
                database_name = key
                break

        if database_name is None:
            return None

        database_meta = existing_databases_meta[database_name]
        if database_meta["type"] == "integration":
            integration = self.integration_controller.get(name=database_name)
            return IntegrationDataNode(
                database_name,
                ds_type=integration["engine"],
                integration_controller=self.session.integration_controller,
            )
        if database_meta["type"] == "project":
            project = self.database_controller.get_project(name=database_name)
            return ProjectDataNode(
                project=project,
                integration_controller=self.session.integration_controller,
                information_schema=self,
            )

        integration_names = self.integration_controller.get_all().keys()
        for integration_name in integration_names:
            if integration_name.lower() == name_lower:
                datasource = self.integration_controller.get(name=integration_name)
                return IntegrationDataNode(
                    integration_name,
                    ds_type=datasource["engine"],
                    integration_controller=self.session.integration_controller,
                )

        return None

    def has_table(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return True
        return False

    def get_table_columns(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return self.information_schema[tn]
        raise exc.ErTableExistError(
            f"Table information_schema.{tableName} does not exists"
        )

    def get_integrations_names(self):
        integration_names = self.integration_controller.get_all().keys()
        # remove files from list to prevent doubling in 'select from INFORMATION_SCHEMA.TABLES'
        return [x.lower() for x in integration_names if x not in ("files",)]

    def get_projects_names(self):
        projects = self.database_controller.get_dict(filter_type="project")
        return [x.lower() for x in projects]

    def _get_handlers(self, query: ASTNode = None):
        columns = self.information_schema["HANDLERS"]

        handlers = self.integration_controller.get_handlers_import_status()

        data = []
        for _key, val in handlers.items():
            connection_args = val.get("connection_args")
            if connection_args is not None:
                connection_args = str(dict(connection_args))
            import_success = val.get("import", {}).get("success")
            import_error = val.get("import", {}).get("error_message")
            data.append(
                [
                    val["name"],
                    val.get("type"),
                    val.get("title"),
                    val.get("description"),
                    val.get("version"),
                    connection_args,
                    import_success,
                    import_error,
                ]
            )

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_ml_engines(self, query: ASTNode = None):
        columns = self.information_schema["ML_ENGINES"]

        integrations = self.integration_controller.get_all()
        ml_integrations = {
            key: val for key, val in integrations.items() if val["type"] == "ml"
        }

        data = []
        for _key, val in ml_integrations.items():
            data.append([val["name"], val.get("engine"), val.get("connection_data")])

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_tables(self, query: ASTNode = None):
        columns = self.information_schema["TABLES"]

        target_table = None
        if (
            type(query) == Select
            and type(query.where) == BinaryOperation
            and query.where.op == "and"
        ):
            for arg in query.where.args:
                if (
                    type(arg) == BinaryOperation
                    and arg.op == "="
                    and type(arg.args[0]) == Identifier
                    and arg.args[0].parts[-1].upper() == "TABLE_SCHEMA"
                    and type(arg.args[1]) == Constant
                ):
                    target_table = arg.args[1].value
                    break

        data = []
        for name in self.information_schema.keys():
            if target_table is not None and target_table != name:
                continue
            row = TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.SYSTEM_VIEW, TABLE_NAME=name)
            data.append(row.to_list())

        for ds_name, ds in self.persis_datanodes.items():
            if target_table is not None and target_table != ds_name:
                continue
            ds_tables = ds.get_tables()
            if len(ds_tables) == 0:
                continue
            elif isinstance(ds_tables[0], dict):
                ds_tables = [
                    TablesRow(
                        TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x["name"]
                    )
                    for x in ds_tables
                ]
            elif (
                isinstance(ds_tables, list)
                and len(ds_tables) > 0
                and isinstance(ds_tables[0], str)
            ):
                ds_tables = [
                    TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x)
                    for x in ds_tables
                ]
            for row in ds_tables:
                row.TABLE_SCHEMA = ds_name
                data.append(row.to_list())

        for ds_name in self.get_integrations_names():
            if target_table is not None and target_table != ds_name:
                continue
            try:
                ds = self.get(ds_name)
                ds_tables = ds.get_tables()
                for row in ds_tables:
                    row.TABLE_SCHEMA = ds_name
                    data.append(row.to_list())
            except Exception:
                logger.error(f"Can't get tables from '{ds_name}'")

        for project_name in self.get_projects_names():
            if target_table is not None and target_table != project_name:
                continue
            project_dn = self.get(project_name)
            project_tables = project_dn.get_tables()
            for row in project_tables:
                row.TABLE_SCHEMA = project_name
                data.append(row.to_list())

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_jobs(self, query: ASTNode = None):
        jobs_controller = JobsController()

        project_name = None
        if (
            isinstance(query, Select)
            and type(query.where) == BinaryOperation
            and query.where.op == "="
            and query.where.args[0].parts == ["project"]
            and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        data = jobs_controller.get_list(project_name)

        columns = self.information_schema["JOBS"]
        columns_lower = [col.lower() for col in columns]

        # to list of lists
        data = [[row[k] for k in columns_lower] for row in data]

        return pd.DataFrame(data, columns=columns)

    def _get_jobs_history(self, query: ASTNode = None):
        jobs_controller = JobsController()

        project_name = None
        if (
            isinstance(query, Select)
            and type(query.where) == BinaryOperation
            and query.where.op == "="
            and query.where.args[0].parts == ["project"]
            and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        data = jobs_controller.get_history(project_name)

        columns = self.information_schema["JOBS_HISTORY"]
        columns_lower = [col.lower() for col in columns]

        # to list of lists
        data = [[row[k] for k in columns_lower] for row in data]

        return pd.DataFrame(data, columns=columns)

    def _get_triggers(self, query: ASTNode = None):
        from mindsdb.interfaces.triggers.triggers_controller import TriggersController

        triggers_controller = TriggersController()

        project_name = None
        if (
            isinstance(query, Select)
            and type(query.where) == BinaryOperation
            and query.where.op == "="
            and query.where.args[0].parts == ["project"]
            and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        data = triggers_controller.get_list(project_name)

        columns = self.information_schema["MDB_TRIGGERS"]
        columns_lower = [col.lower() for col in columns]

        # to list of lists
        data = [[row[k] for k in columns_lower] for row in data]

        return pd.DataFrame(data, columns=columns)

    def _get_chatbots(self, query: ASTNode = None):
        from mindsdb.interfaces.chatbot.chatbot_controller import ChatBotController

        chatbot_controller = ChatBotController()

        project_name = None
        if (
            isinstance(query, Select)
            and type(query.where) == BinaryOperation
            and query.where.op == "="
            and query.where.args[0].parts == ["project"]
            and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        data = chatbot_controller.get_chatbots(project_name)

        columns = self.information_schema["CHATBOTS"]
        columns_lower = [col.lower() for col in columns]

        # to list of lists
        data = [[row[k] for k in columns_lower] for row in data]

        return pd.DataFrame(data, columns=columns)

    def _get_knowledge_bases(self, query: ASTNode = None):
        from mindsdb.interfaces.knowledge_base.controller import KnowledgeBaseController
        controller = KnowledgeBaseController(self.session)
        project_name = None
        if (
                isinstance(query, Select)
                and type(query.where) == BinaryOperation
                and query.where.op == '='
                and query.where.args[0].parts == ['project']
                and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        # get the project id from the project name
        project_controller = ProjectController()
        project_id = project_controller.get(name=project_name).id
        kb_list = controller.list(project_id=project_id)

        columns = self.information_schema['KNOWLEDGE_BASES']

        # columns: NAME, PROJECT, MODEL, STORAGE
        data = [
            (
                kb.name,
                project_name,
                kb.embedding_model.name,
                kb.vector_database.name + '.' + kb.vector_database_table
            ) for kb in kb_list
        ]

        return pd.DataFrame(data, columns=columns)

    def _get_skills(self, query: ASTNode = None):
        skills_controller = SkillsController()
        project_name = None
        if (
                isinstance(query, Select)
                and type(query.where) == BinaryOperation
                and query.where.op == '='
                and query.where.args[0].parts == ['project']
                and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        all_skills = skills_controller.get_skills(project_name)

        columns = self.information_schema['SKILLS']

        # NAME, PROJECT, TYPE, PARAMS
        data = [(s.name, project_name, s.type, s.params) for s in all_skills]
        return pd.DataFrame(data, columns=columns)

    def _get_agents(self, query: ASTNode = None):
        agents_controller = AgentsController()
        project_name = None
        if (
                isinstance(query, Select)
                and type(query.where) == BinaryOperation
                and query.where.op == '='
                and query.where.args[0].parts == ['project']
                and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        all_agents = agents_controller.get_agents(project_name)

        columns = self.information_schema['AGENTS']

        # NAME, PROJECT, MODEL, SKILLS, PARAMS
        data = [(a.name, project_name, a.model_name, list(map(lambda s: s.name, a.skills)), a.params) for a in all_agents]
        return pd.DataFrame(data, columns=columns)

    def _get_databases(self, query: ASTNode = None):
        columns = self.information_schema["DATABASES"]

        project = self.database_controller.get_list()
        data = [
            [x["name"], x["type"], x["engine"], str(x.get("connection_data"))]
            for x in project
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_models(self, query: ASTNode = None):
        columns = self.information_schema["MODELS"]
        data = []
        for project_name in self.get_projects_names():
            project = self.database_controller.get_project(name=project_name)
            project_models = project.get_models()
            for row in project_models:
                table_name = row["name"]
                table_meta = row["metadata"]
                if table_meta["active"] is not True:
                    continue
                data.append(
                    [
                        table_name,
                        table_meta["engine"],
                        project_name,
                        table_meta["version"],
                        table_meta["status"],
                        table_meta["accuracy"],
                        table_meta["predict"],
                        table_meta["update_status"],
                        table_meta["mindsdb_version"],
                        table_meta["error"],
                        table_meta["select_data_query"],
                        table_meta["training_options"],
                        table_meta["current_training_phase"],
                        table_meta["total_training_phases"],
                        table_meta["training_phase_name"],
                        table_meta["label"],
                        row["created_at"],
                        table_meta["training_time"],
                    ]
                )
            # TODO optimise here
            # if target_table is not None and target_table != project_name:
            #     continue

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_models_versions(self, query: ASTNode = None):
        columns = self.information_schema["MODELS_VERSIONS"]
        data = []
        for project_name in self.get_projects_names():
            project = self.database_controller.get_project(name=project_name)
            project_models = project.get_models(active=None)
            for row in project_models:
                table_name = row["name"]
                table_meta = row["metadata"]
                data.append(
                    [
                        table_name,
                        table_meta["engine"],
                        project_name,
                        table_meta["active"],
                        table_meta["version"],
                        table_meta["status"],
                        table_meta["accuracy"],
                        table_meta["predict"],
                        table_meta["update_status"],
                        table_meta["mindsdb_version"],
                        table_meta["error"],
                        table_meta["select_data_query"],
                        table_meta["training_options"],
                        table_meta["label"],
                        row["created_at"],
                        table_meta["training_time"],
                    ]
                )

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_columns(self, query: ASTNode = None):
        columns = self.information_schema["COLUMNS"]

        # NOTE there is a lot of types in mysql, but listed below should be enough for our purposes
        row_templates = {
            "text": [
                "def",
                "SCHEMA_NAME",
                "TABLE_NAME",
                "COLUMN_NAME",
                "COL_INDEX",
                None,
                "YES",
                "varchar",
                1024,
                3072,
                None,
                None,
                None,
                "utf8",
                "utf8_bin",
                "varchar(1024)",
                None,
                None,
                "select",
                None,
                None,
            ],
            "timestamp": [
                "def",
                "SCHEMA_NAME",
                "TABLE_NAME",
                "COLUMN_NAME",
                "COL_INDEX",
                "CURRENT_TIMESTAMP",
                "YES",
                "timestamp",
                None,
                None,
                None,
                None,
                0,
                None,
                None,
                "timestamp",
                None,
                None,
                "select",
                None,
                None,
            ],
            "bigint": [
                "def",
                "SCHEMA_NAME",
                "TABLE_NAME",
                "COLUMN_NAME",
                "COL_INDEX",
                None,
                "YES",
                "bigint",
                None,
                None,
                20,
                0,
                None,
                None,
                None,
                "bigint unsigned",
                None,
                None,
                "select",
                None,
                None,
            ],
            "float": [
                "def",
                "SCHEMA_NAME",
                "TABLE_NAME",
                "COLUMN_NAME",
                "COL_INDEX",
                None,
                "YES",
                "float",
                None,
                None,
                12,
                0,
                None,
                None,
                None,
                "float",
                None,
                None,
                "select",
                None,
                None,
            ],
        }

        result = []

        for table_name in self.information_schema:
            table_columns = self.information_schema[table_name]
            for i, column_name in enumerate(table_columns):
                result_row = row_templates["text"].copy()
                result_row[1] = "information_schema"
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        mindsdb_dn = self.get("MINDSDB")
        for table_row in mindsdb_dn.get_tables():
            table_name = table_row.TABLE_NAME
            table_columns = mindsdb_dn.get_table_columns(table_name)
            for i, column_name in enumerate(table_columns):
                result_row = row_templates["text"].copy()
                result_row[1] = "mindsdb"
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        files_dn = self.get("FILES")
        for table_name in files_dn.get_tables():
            table_columns = files_dn.get_table_columns(table_name)
            for i, column_name in enumerate(table_columns):
                result_row = row_templates["text"].copy()
                result_row[1] = "files"
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        df = pd.DataFrame(result, columns=columns)
        return df

    def _get_schemata(self, query: ASTNode = None):
        columns = self.information_schema["SCHEMATA"]

        databases_meta = self.session.database_controller.get_list()
        data = [
            ["def", x["name"], "utf8mb4", "utf8mb4_0900_ai_ci", None]
            for x in databases_meta
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_engines(self, query: ASTNode = None):
        columns = self.information_schema["ENGINES"]
        data = [
            [
                "InnoDB",
                "DEFAULT",
                "Supports transactions, row-level locking, and foreign keys",
                "YES",
                "YES",
                "YES",
            ]
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_charsets(self, query: ASTNode = None):
        columns = self.information_schema["CHARACTER_SETS"]
        data = [
            ["utf8", "UTF-8 Unicode", "utf8_general_ci", 3],
            ["latin1", "cp1252 West European", "latin1_swedish_ci", 1],
            ["utf8mb4", "UTF-8 Unicode", "utf8mb4_general_ci", 4],
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_collations(self, query: ASTNode = None):
        columns = self.information_schema["COLLATIONS"]
        data = [
            ["utf8_general_ci", "utf8", 33, "Yes", "Yes", 1, "PAD SPACE"],
            ["latin1_swedish_ci", "latin1", 8, "Yes", "Yes", 1, "PAD SPACE"],
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_empty_table(self, table_name, query: ASTNode = None):
        columns = self.information_schema[table_name]
        data = []

        df = pd.DataFrame(data, columns=columns)
        return df

    def query(self, query: ASTNode, session=None):
        query_tables = get_all_tables(query)

        if len(query_tables) != 1:
            raise exc.ErBadTableError(
                f"Only one table can be used in query to information_schema: {query}"
            )

        table_name = query_tables[0].upper()

        if table_name not in self.get_dataframe_funcs:
            raise exc.ErNotSupportedYet("Information schema: Not implemented.")

        dataframe = self.get_dataframe_funcs[table_name](query=query)
        data = query_df(dataframe, query, session=self.session)

        columns_info = [{"name": k, "type": v} for k, v in data.dtypes.items()]

        return data.to_dict(orient="records"), columns_info



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import (
    DEFAULT_AUTH_METHOD,
    DEFAULT_COALLITION_ID,
    FILLER_FOR_WIRESHARK_DUMP,
    SERVER_STATUS_AUTOCOMMIT
)
from mindsdb.api.mysql.mysql_proxy.classes.server_capabilities import server_capabilities
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum


class HandshakePacket(Packet):
    '''
    Implementation based on:
    https://mariadb.com/kb/en/library/1-connecting-connecting/#initial-handshake-packet
    '''

    def setup(self):
        capabilities = server_capabilities.value
        self.protocol_version = Datum('int<1>', 10)
        self.server_version = Datum('string<NUL>', '5.7.1-MindsDB-1.0')
        self.connection_id = Datum('int<4>', self.proxy.connection_id)
        self.scramble_1st_part = Datum('string<8>', self.proxy.salt[:8])
        self.reserved_byte = Datum('string<1>', '')
        self.server_capabilities_1st_part = Datum('int<2>', capabilities)
        self.server_default_collation = Datum('int<1>', DEFAULT_COALLITION_ID)
        self.status_flags = Datum('int<2>', SERVER_STATUS_AUTOCOMMIT)
        self.server_capabilities_2nd_part = Datum('int<2>', capabilities >> 16)
        self.wireshark_filler = Datum('int<1>', FILLER_FOR_WIRESHARK_DUMP)
        # self.wireshark_filler = Datum('int<1>', len(self.proxy.salt))
        self.reserved_filler1 = Datum('string<6>', '')
        self.reserved_filler2 = Datum('string<4>', '')
        self.scramble_2nd_part = Datum('string<NUL>', self.proxy.salt[8:])
        self.null_close = Datum('string<NUL>', DEFAULT_AUTH_METHOD)

    @property
    def body(self):
        order = [
            'protocol_version',
            'server_version',
            'connection_id',
            'scramble_1st_part',
            'reserved_byte',
            'server_capabilities_1st_part',
            'server_default_collation',
            'status_flags',
            'server_capabilities_2nd_part',
            'wireshark_filler',
            'reserved_filler1',
            'reserved_filler2',
            'scramble_2nd_part',
            'null_close'
        ]
        string = b''
        for key in order:
            string += getattr(self, key).toStringPacket()

        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint
        pprint.pprint(str(HandshakePacket().get_packet_string()))


# only run the test if this file is called from debugger
if __name__ == "__main__":
    HandshakePacket.test()



import traceback

from flask import request
from flask_restx import Resource

import mindsdb.utilities.hooks as hooks
import mindsdb.utilities.profiler as profiler
from mindsdb.api.http.namespaces.configs.sql import ns_conf
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import (
    RESPONSE_TYPE as SQL_RESPONSE_TYPE,
)
from mindsdb.api.mysql.mysql_proxy.utilities import SqlApiException, SqlApiUnknownError
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx

logger = log.getLogger(__name__)


@ns_conf.route("/query")
@ns_conf.param("query", "Execute query")
class Query(Resource):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @ns_conf.doc("query")
    def post(self):
        query = request.json["query"]
        context = request.json.get("context", {})

        if context.get("profiling") is True:
            profiler.enable()

        error_type = None
        error_code = None
        error_text = None
        error_traceback = None

        profiler.set_meta(
            query=query, api="http", environment=Config().get("environment")
        )
        with profiler.Context("http_query_processing"):
            mysql_proxy = FakeMysqlProxy()
            mysql_proxy.set_context(context)
            try:
                result = mysql_proxy.process_query(query)

                if result.type == SQL_RESPONSE_TYPE.OK:
                    query_response = {"type": SQL_RESPONSE_TYPE.OK}
                elif result.type == SQL_RESPONSE_TYPE.TABLE:
                    query_response = {
                        "type": SQL_RESPONSE_TYPE.TABLE,
                        "data": result.data,
                        "column_names": [
                            x["alias"] or x["name"] if "alias" in x else x["name"]
                            for x in result.columns
                        ],
                    }
            except SqlApiException as e:
                # classified error
                error_type = "expected"
                query_response = {
                    "type": SQL_RESPONSE_TYPE.ERROR,
                    "error_code": e.err_code,
                    "error_message": str(e),
                }

            except SqlApiUnknownError as e:
                # unclassified
                error_type = "unexpected"
                query_response = {
                    "type": SQL_RESPONSE_TYPE.ERROR,
                    "error_code": e.err_code,
                    "error_message": str(e),
                }

            except Exception as e:
                error_type = "unexpected"
                query_response = {
                    "type": SQL_RESPONSE_TYPE.ERROR,
                    "error_code": 0,
                    "error_message": str(e),
                }
                logger.error(f"Error profiling query: \n{traceback.format_exc()}")

            if query_response.get("type") == SQL_RESPONSE_TYPE.ERROR:
                error_type = "expected"
                error_code = query_response.get("error_code")
                error_text = query_response.get("error_message")

            context = mysql_proxy.get_context(context)

            query_response["context"] = context

        hooks.after_api_query(
            company_id=ctx.company_id,
            api="http",
            command=None,
            payload=query,
            error_type=error_type,
            error_code=error_code,
            error_text=error_text,
            traceback=error_traceback,
        )

        return query_response, 200


@ns_conf.route("/list_databases")
@ns_conf.param("list_databases", "lists databases of mindsdb")
class ListDatabases(Resource):
    @ns_conf.doc("list_databases")
    def get(self):
        listing_query = "SHOW DATABASES"
        mysql_proxy = FakeMysqlProxy()
        try:
            result = mysql_proxy.process_query(listing_query)

            # iterate over result.data and perform a query on each item to get the name of the tables
            if result.type == SQL_RESPONSE_TYPE.ERROR:
                listing_query_response = {
                    "type": "error",
                    "error_code": result.error_code,
                    "error_message": result.error_message,
                }
            elif result.type == SQL_RESPONSE_TYPE.OK:
                listing_query_response = {"type": "ok"}
            elif result.type == SQL_RESPONSE_TYPE.TABLE:
                listing_query_response = {
                    "data": [
                        {
                            "name": x[0],
                            "tables": mysql_proxy.process_query(
                                "SHOW TABLES FROM `{}`".format(x[0])
                            ).data,
                        }
                        for x in result.data
                    ]
                }
        except Exception as e:
            listing_query_response = {
                "type": "error",
                "error_code": 0,
                "error_message": str(e),
            }

        return listing_query_response, 200



"""chatbot_to_task

Revision ID: ad04ee0bd385
Revises: a57506731839
Create Date: 2023-07-19 16:18:14.791761

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa

# revision identifiers, used by Alembic.
revision = 'ad04ee0bd385'
down_revision = 'a57506731839'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('chat_bots', schema=None) as batch_op:
        batch_op.drop_column('chat_engine')
        batch_op.drop_column('user_class')
        batch_op.drop_column('company_id')
        batch_op.drop_column('is_running')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('chat_bots', schema=None) as batch_op:
        batch_op.add_column(sa.Column('is_running', sa.BOOLEAN(), nullable=True))
        batch_op.add_column(sa.Column('company_id', sa.INTEGER(), nullable=True))
        batch_op.add_column(sa.Column('user_class', sa.INTEGER(), nullable=True))
        batch_op.add_column(sa.Column('chat_engine', sa.VARCHAR(), nullable=True))

    # ### end Alembic commands ###



import pandas as pd
from typing import Text, List, Dict, Any

from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.handlers.utilities.query_utilities.insert_query_utilities import INSERTQueryParser

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MessagesTable(APITable):
    """The Microsoft Teams Messages Table implementation"""

    def insert(self, query: ast.Insert) -> None:
        """Sends messages to a Microsoft Teams Channel through a configured webhook.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['text', 'title'],
            mandatory_columns=['text'],
            all_mandatory=False
        )
        message_data = insert_statement_parser.parse_query()
        self.send_message(message_data)

    def send_message(self, message_data: List[Dict[Text, Any]]) -> None:
        """Sends messages to a Microsoft Teams Channel using the parsed message data.

        Parameters
        ----------
        message_data : List[Dict[Text, Any]]
           List of dictionaries containing the messages to send.

        Returns
        -------
        None
        """
        teams = self.handler.connect() 
        for message in message_data:
            try:
                teams.title(message['title'])
                teams.text(message['text'])
                teams.send()
                logger.info(f"Message sent to Microsoft Teams channel successfully.")
            except Exception as e:
                logger.error(f"Error sending message to Microsoft Teams channel: {e}")
                raise e


"""
    Wrapper around MindsDB's executor and integration controller following the implementation of the original
    langchain.sql_database.SQLDatabase class to partly replicate its behavior.
"""
import warnings
from typing import Iterable, List, Optional
from mindsdb_sql import parse_sql
from langchain.sql_database import SQLDatabase

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MindsDBSQL(SQLDatabase):
    """ Can't modify signature, as LangChain does a Pydantic check."""
    def __init__(
        self,
        engine,
        schema: Optional[str] = None,
        metadata: Optional = None,
        ignore_tables: Optional[List[str]] = None,
        include_tables: Optional[List[str]] = None,
        sample_rows_in_table_info: int = 3,
        indexes_in_table_info: bool = False,
        custom_table_info: Optional[dict] = None,
        view_support: Optional[bool] = True,
    ):
        # Some args above are not used in this class, but are kept for compatibility
        self._engine = engine   # executor instance
        self._metadata = metadata  # integrations controller instance
        self._sample_rows_in_table_info = int(sample_rows_in_table_info)
        
        self._tables_to_include = include_tables
        self._tables_to_ignore = []
        if not self._tables_to_include:
            # ignore_tables and include_tables should not be used together.
            # include_tables takes priority if it's set.
            self._tables_to_ignore = ignore_tables

    @property
    def dialect(self) -> str:
        return 'mindsdb'

    def _call_engine(self, queries: List[str]):
        for query in queries:
            self._engine.is_executed = False
            ast_query = parse_sql(query.strip('`'), dialect='mindsdb')
            ret = self._engine.execute_command(ast_query)
        return ret

    def get_usable_table_names(self) -> Iterable[str]:
        if self._tables_to_include:
            return self._tables_to_include
        
        original_db = self._engine.session.database
        ret = self._call_engine(['show databases;'])
        dbs = [lst[0] for lst in ret.data if lst[0] != 'information_schema']
        usable_tables = []
        for db in dbs:
            if db != 'mindsdb':
                try:
                    ret = self._call_engine([f'use `{db}`;', 'show tables;'])
                    tables = [lst[0] for lst in ret.data if lst[0] != 'information_schema']
                    for table in tables:
                        table_name = f'{db}.{table}'
                        if table_name not in self._tables_to_ignore:
                            usable_tables.append(table_name)
                except Exception as e:
                    logger.warning('Unable to get tables for %s: %s', db, str(e))
                finally:
                    _ = self._call_engine([f'use {original_db};'])
        return usable_tables

    def get_table_names(self) -> Iterable[str]:
        warnings.warn("This method is deprecated - please use `get_usable_table_names`.")
        return self.get_usable_table_names()

    @property
    def table_info(self) -> str:
        """Information about all tables in the database."""
        return self.get_table_info()

    def get_table_info(self, table_names: Optional[List[str]] = None) -> str:
        """ Get information about specified tables.
        Follows best practices as specified in: Rajkumar et al, 2022 (https://arxiv.org/abs/2204.00498)
        If `sample_rows_in_table_info`, the specified number of sample rows will be
        appended to each table description. This can increase performance as demonstrated in the paper.
        """
        all_table_names = self.get_usable_table_names()
        if table_names is not None:
            missing_tables = set(table_names).difference(all_table_names)
            if missing_tables:
                raise ValueError(f"table_names {missing_tables} not found in database")
            all_table_names = table_names

        tables = []
        for table in all_table_names:
            table_info = self._get_single_table_info(table)
            tables.append(table_info)

        final_str = "\n\n".join(tables)
        return final_str

    def _get_single_table_info(self, table_str: str) -> str:
        controller = self._metadata
        integration, table_name = table_str.split('.')
        cols_df = controller.get_handler(integration).get_columns(table_name).data_frame
        fields = cols_df['Field'].to_list()
        dtypes = cols_df['Type'].to_list()

        info = f'Table named `{table_name}`\n'
        info += f"\n/* Sample with first {self._sample_rows_in_table_info} rows from table `{table_str}`:\n"
        info += "\t".join([field for field in fields])
        info += self._get_sample_rows(table_str, fields) + "\n*/"
        info += '\nColumn data types: ' + ",\t".join([f'`{field}` : `{dtype}`' for field, dtype in zip(fields, dtypes)]) + '\n'  # noqa
        return info

    def _get_sample_rows(self, table: str, fields: List[str]) -> str:
        command = f"select {','.join(fields)} from {table} limit {self._sample_rows_in_table_info};"
        try:
            ret = self._call_engine([command])
            sample_rows = ret.data
            sample_rows = list(map(lambda ls: [str(i) if len(str(i)) < 100 else str[:100]+'...' for i in ls], sample_rows))
            sample_rows_str = "\n" + "\n".join(["\t".join(row) for row in sample_rows])
        except Exception:
            sample_rows_str = "\n" + "\t [error] Couldn't retrieve sample rows!"

        return sample_rows_str

    def run(self, command: str, fetch: str = "all") -> str:
        """Execute a SQL command and return a string representing the results.
        If the statement returns rows, a string of the results is returned.
        If the statement returns no rows, an empty string is returned.
        """
        def _tidy(result: List) -> str:
            return '\n'.join(['\t'.join([str(value) for value in row]) for row in result])
        ret = self._call_engine([command])
        if fetch == "all":
            result = _tidy(ret.data)
        elif fetch == "one":
            result = _tidy(ret.data[0])
        else:
            raise ValueError("Fetch parameter must be either 'one' or 'all'")
        return str(result)

    def get_table_info_no_throw(self, table_names: Optional[List[str]] = None) -> str:
        try:
            return self.get_table_info(table_names)
        except Exception as e:
            return f"Error: {e}"

    def run_no_throw(self, command: str, fetch: str = "all") -> str:
        try:
            return self.run(command, fetch)
        except Exception as e:
            return f"Error: {e}"



from collections import OrderedDict

import pandas as pd
import pyodbc

from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb_sql import parse_sql
from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

logger = log.getLogger(__name__)

class HSQLDBHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the HyperSQL statements.
    """

    name = 'hsqldb'

    def __init__(self, name: str,  **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'hsqldb'
        self.connection_args = kwargs.get('connection_data')
        self.server_name = self.connection_args.get('server_name', 'localhost')
        self.port = self.connection_args.get('port')
        self.database_name = self.connection_args.get('database_name')
        self.username = self.connection_args.get('username')
        self.password = self.connection_args.get('password')
        self.conn_str = f"DRIVER={{PostgreSQL Unicode}};SERVER={self.server_name};PORT={self.port};DATABASE={self.database_name};UID={self.username};PWD={self.password};Trusted_Connection=True"
        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = pyodbc.connect(self.conn_str, timeout=10)
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to SQLite, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cursor:
            try:
                cursor.execute(query)
                result = cursor.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame.from_records(
                            result,
                            columns=[x[0] for x in cursor.description]
                        )
                    )

                else:
                    response = Response(RESPONSE_TYPE.OK)
                    connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        renderer = SqlalchemyRender('postgres')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        connection = self.connect()
        cursor = connection.cursor()
        cursor.execute("SELECT * FROM information_schema.tables WHERE table_schema NOT IN ('information_schema', 'pg_catalog') AND table_type='BASE TABLE'")
        results = cursor.fetchall()
        df = pd.DataFrame([x[2] for x in results], columns=['table_name']) # Workaround since cursor.tables() wont work with postgres driver
        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )

        return response

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        connection = self.connect()
        cursor = connection.cursor()
        query = f'SELECT * FROM information_schema.columns WHERE table_name ={table_name}' # Workaround since cursor.columns() wont work with postgres driver
        cursor.execute(query)
        results = cursor.fetchall()
        df = pd.DataFrame(
            [(x[3], x[7]) for x in results],
            columns=['column_name', 'data_type']
        )

        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )

        return response

connection_args = OrderedDict(
    server_name={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the database'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect.'
    },
    database_name={
        'type': ARG_TYPE.STR,
        'description': '''
            The database name to use when connecting.
        '''
    },
    username={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate.'
    },
    
)

connection_args_example = OrderedDict(
    server_name = 'localhost',
    port =  9001,
    database_name = 'xdb',
    username = 'SA',
    password = 'password'
)


import unittest
from mindsdb.integrations.handlers.crate_handler.crate_handler import CrateHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class CrateHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "127.0.0.1",
            "port": 4200,
            "user": "crate",
            "password": "",
        }
        cls.handler = CrateHandler("test_crate_handler", cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_drop_table(self):
        res = self.handler.query("DROP TABLE IF EXISTS PREM;")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_2_create_table(self):
        res = self.handler.query("CREATE TABLE IF NOT EXISTS PREM (Premi varchar(50));")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_3_insert_table(self):
        res = self.handler.query("INSERT INTO PREM VALUES('Radha <3 Krishna');")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_4_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_5_select_query(self):
        query = "SELECT * FROM PREM;"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE or RESPONSE_TYPE.OK

    def test_6_check_connection(self):
        self.handler.check_connection()


if __name__ == "__main__":
    unittest.main()



import concurrent.futures
import io
import re
import traceback
from threading import Lock
from urllib.parse import urljoin, urlparse

import fitz  # PyMuPDF
import pandas as pd
import requests
from bs4 import BeautifulSoup
from mindsdb.utilities import log

logger = log.getLogger(__name__)

def pdf_to_markdown(response):
    # Download the PDF from the given URL

    file_stream = io.BytesIO(response.content)

    # Open the PDF from the in-memory file
    document = fitz.open(stream=file_stream, filetype="pdf")

    markdown_text = ""
    for page_num in range(len(document)):
        page = document.load_page(page_num)

        # Get the blocks of text
        blocks = page.get_text("blocks")

        # Sort the blocks by their vertical position on the page
        blocks.sort(key=lambda block: (block[1], block[0]))

        previous_block_bottom = 0
        for block in blocks:
            y0 = block[1]
            y1 = block[3]
            block_text = block[4]

            # Check if there's a large vertical gap between this block and the previous one
            if y0 - previous_block_bottom > 10:
                markdown_text += "\n"

            markdown_text += block_text + "\n"
            previous_block_bottom = y1

        markdown_text += "\n"

    # Close the document
    document.close()

    return markdown_text


url_list_lock = Lock()


def is_valid(url):
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)


# this bad boy gets all the crawling done in parallel
def parallel_get_all_website_links(urls):
    url_contents = {}

    if len(urls) <= 10:
        for url in urls:
            url_contents[url] = get_all_website_links(url)
        return url_contents

    with concurrent.futures.ProcessPoolExecutor() as executor:
        future_to_url = {
            executor.submit(get_all_website_links, url): url for url in urls
        }
        for future in concurrent.futures.as_completed(future_to_url):
            url = future_to_url[future]
            try:
                url_contents[url] = future.result()
            except Exception as exc:
                logger.error(f'{url} generated an exception: {exc}')
   
    return url_contents


# this crawls one individual website
def get_all_website_links(url):
    logger.info("crawling: {url} ...".format(url=url))
    urls = set()

    domain_name = urlparse(url).netloc
    try:
        # Create a session to handle cookies
        session = requests.Session()

        # Add headers to mimic a real browser request
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36"
        }

        # Send GET request
        response = session.get(url, headers=headers)
        # Accept cookies if necessary
        if "cookie" in response.request.headers:
            session.cookies.update(response.cookies)

        content_type = response.headers.get("Content-Type", "").lower()

        if "application/pdf" in content_type:
            content_html = "PDF"
            content_text = pdf_to_markdown(response)
        else:
            content_html = response.text

            # Parse HTML content with BeautifulSoup
            soup = BeautifulSoup(content_html, "html.parser")
            content_text = get_readable_text_from_soup(soup)
            for a_tag in soup.findAll("a"):
                href = a_tag.attrs.get("href")
                if href == "" or href is None:
                    continue
                href = urljoin(url, href)
                parsed_href = urlparse(href)

                href = (
                    parsed_href.scheme + "://" + parsed_href.netloc + parsed_href.path
                )
                if not is_valid(href):
                    continue
                if href in urls:
                    continue
                if domain_name != parsed_href.netloc:
                    continue

                href = href.rstrip("/")
                urls.add(href)

    except Exception as e:
        error_message = traceback.format_exc().splitlines()[-1]
        logger.error("An exception occurred: %s", str(e))
        return {
            "url": url,
            "urls": urls,
            "html_content": "",
            "text_content": "",
            "error": str(error_message),
        }

    return {
        "url": url,
        "urls": urls,
        "html_content": content_html,
        "text_content": content_text,
        "error": None,
    }


def get_readable_text_from_soup(soup):
    # Start formatting as Markdown
    markdown_output = ""

    # Iterate through headings and paragraphs
    for tag in soup.find_all(
        ["h1", "h2", "h3", "h4", "h5", "h6", "p", "a", "ul", "ol", "li"]
    ):
        if tag.name in ["h1", "h2", "h3", "h4", "h5", "h6"]:
            markdown_output += (
                "#" * int(tag.name[1]) + " " + tag.get_text().strip() + "\n\n"
            )
        elif tag.name == "p":
            markdown_output += tag.get_text().strip() + "\n\n"
        elif tag.name == "a":
            markdown_output += f"[{tag.get_text().strip()}]({tag.get('href')})\n\n"
        elif tag.name == "ul":
            for li in tag.find_all("li"):
                markdown_output += f"* {li.get_text().strip()}\n"
            markdown_output += "\n"
        elif tag.name == "ol":
            for index, li in enumerate(tag.find_all("li")):
                markdown_output += f"{index + 1}. {li.get_text().strip()}\n"
            markdown_output += "\n"

    return markdown_output


# this bad girl does the recursive crawling of the websites
def get_all_website_links_rec(url, reviewd_urls, limit=None):
    if limit is not None:
        if len(reviewd_urls) >= limit:
            return reviewd_urls

    if url not in reviewd_urls:
        # if something happens getting the website links for this url then log the error
        try:
            reviewd_urls[url] = get_all_website_links(url)
        except Exception as e:
            error_message = traceback.format_exc().splitlines()[-1]
            logger.error("An exception occurred: %s", str(e))
            reviewd_urls[url] = {
                "url": url,
                "urls": [],
                "html_content": "",
                "text_content": "",
                "error": str(error_message),
            }

    to_rev_url_list = []

    # create a list of new urls to review that don't exist in the already reviewed ones
    for new_url in reviewd_urls[url]["urls"]:
        # if this is already in the urls, then no need to go and crawl for it
        if new_url in reviewd_urls or new_url in to_rev_url_list:
            continue

        # insert immediately to count limit between threads. fill later
        with url_list_lock:
            if limit is None or len(reviewd_urls) < limit:
                reviewd_urls[new_url] = {}
                to_rev_url_list.append(new_url)
            else:
                break

    # if there is something to fetch, go fetch
    if len(to_rev_url_list) > 0:
        new_revised_urls = parallel_get_all_website_links(to_rev_url_list)

        reviewd_urls.update(new_revised_urls)

        for new_url in new_revised_urls:
            get_all_website_links_rec(new_url, reviewd_urls, limit)


# this crawls the websites and returns it all as a dataframe, ready to be served
def get_all_websites(urls, limit=1, html=False):
    reviewd_urls = {}

    # def fetch_url(url):
    #     url = url.rstrip('/')
    #     if urlparse(url).scheme == "":
    #         # Try HTTPS first
    #         url = "https://" + url
    #     reviewd_urls_iter = {}
    #     get_all_website_links_rec(url, reviewd_urls_iter, limit)
    #     return reviewd_urls_iter

    # reviewd_urls = fetch_url(urls[0])
    # Define a helper function that will be run in parallel.
    def fetch_url(url):
        # Allow URLs to be passed wrapped in quotation marks so they can be used
        # directly from the SQL editor.
        if url.startswith("'") and url.endswith("'"):
            url = url[1:-1]
        url = url.rstrip("/")
        if urlparse(url).scheme == "":
            # Try HTTPS first
            url = "https://" + url
        get_all_website_links_rec(url, reviewd_urls, limit)

    # Use a ThreadPoolExecutor to run the helper function in parallel.
    with concurrent.futures.ThreadPoolExecutor() as executor:
        future_to_url = {executor.submit(fetch_url, url): url for url in urls}

        for future in concurrent.futures.as_completed(future_to_url):
            future.result()

    columns_to_ignore = ["urls"]
    if html is False:
        columns_to_ignore += ["html_content"]
    df = dict_to_dataframe(
        reviewd_urls, columns_to_ignore=columns_to_ignore, index_name="url"
    )

    if not df.empty and df[df.error.isna()].empty:
        # no real data - rise exception from first row
        raise Exception(str(df.iloc[0].error))
    return df


# this can parse the native query
def parse_urls_limit(input_str):
    # Split the input string into 'url', 'limit' or 'html' parts
    items = re.split(r",", input_str)

    # Initialize list for urls, limit and html
    urls = []
    limit = None
    html = False

    for item in items:
        item = item.strip()  # Remove leading/trailing whitespace

        # Check if item is a 'limit' or 'html' setting
        if item.lower().startswith("limit"):
            limit_match = re.search(r"\d+", item)
            if limit_match:
                limit = int(limit_match.group())  # Extract the number
        elif item.lower().startswith("html"):
            html_match = re.search(r"(true|false)", item, re.I)
            if html_match:
                html = (
                    html_match.group().lower() == "true"
                )  # Check if the value is 'true'
        else:
            urls.append(item)  # Add the item to the url list

    return {"urls": urls, "limit": limit, "html": html}


# run a query that goes and crawls urls
# format url, url, ..., limit=n
# you can pass one of many urls, limit is optional
def get_df_from_query_str(query_str):
    args = parse_urls_limit(query_str)
    df = get_all_websites(args["urls"], args["limit"], args["html"])
    return df


# this flips a dictionary of dictionaries into a dataframe so we can use it in mindsdb
def dict_to_dataframe(dict_of_dicts, columns_to_ignore=None, index_name=None):
    # Convert dictionary of dictionaries into DataFrame
    df = pd.DataFrame.from_dict(dict_of_dicts, orient="index")

    # If columns_to_ignore is provided, drop these columns
    if columns_to_ignore:
        df = df.drop(columns_to_ignore, axis=1, errors="ignore")

    # If index_name is provided, rename the index
    if index_name:
        df.index.name = index_name

    return df



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import mindsdb.microservices_grpc.ml.common_pb2 as common__pb2
import mindsdb.microservices_grpc.ml.ml_pb2 as ml__pb2


class MLServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Predict = channel.unary_unary(
                '/MLService/Predict',
                request_serializer=ml__pb2.PredictCall.SerializeToString,
                response_deserializer=common__pb2.Response.FromString,
                )
        self.Create = channel.unary_unary(
                '/MLService/Create',
                request_serializer=ml__pb2.CreateCall.SerializeToString,
                response_deserializer=common__pb2.StatusResponse.FromString,
                )
        self.Update = channel.unary_unary(
                '/MLService/Update',
                request_serializer=ml__pb2.UpdateCall.SerializeToString,
                response_deserializer=common__pb2.StatusResponse.FromString,
                )


class MLServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Predict(self, request, context):
        """A simple RPC.

        Performs 'predict' request and returns the result

        wrapped into Response object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Create(self, request, context):
        """A simple RPC.

        Performs 'learn' request and returns the result

        wrapped into Response object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Update(self, request, context):
        """A simple RPC.

        Performs 'update' request and returns the result

        wrapped into Response object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MLServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Predict': grpc.unary_unary_rpc_method_handler(
                    servicer.Predict,
                    request_deserializer=ml__pb2.PredictCall.FromString,
                    response_serializer=common__pb2.Response.SerializeToString,
            ),
            'Create': grpc.unary_unary_rpc_method_handler(
                    servicer.Create,
                    request_deserializer=ml__pb2.CreateCall.FromString,
                    response_serializer=common__pb2.StatusResponse.SerializeToString,
            ),
            'Update': grpc.unary_unary_rpc_method_handler(
                    servicer.Update,
                    request_deserializer=ml__pb2.UpdateCall.FromString,
                    response_serializer=common__pb2.StatusResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'MLService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class MLService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Predict(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/MLService/Predict',
            ml__pb2.PredictCall.SerializeToString,
            common__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Create(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/MLService/Create',
            ml__pb2.CreateCall.SerializeToString,
            common__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Update(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/MLService/Update',
            ml__pb2.UpdateCall.SerializeToString,
            common__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import pandas as pd
from typing import List
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.handlers.utilities.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.utilities import log
from mindsdb_sql.parser import ast

logger = log.getLogger(__name__)


class ZipCodeBaseCodeLocationTable(APITable):
    """The ZipCodeBase Location Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://app.zipcodebase.com/documentation#search API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Location of the codes matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'code_to_location',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []

        for op, arg1, arg2 in where_conditions:
            if arg1 == "codes":
                if op == '=':
                    search_params["codes"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for codes column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = "codes" in search_params

        if not filter_flag:
            raise NotImplementedError("`codes` column has to be present in where clause.")

        code_to_location_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.client.code_to_location(search_params.get("codes"))

        self.check_res(res=response)

        content = response["content"]

        code_to_location_df = pd.json_normalize(self.clean_resp(content["results"]))

        select_statement_executor = SELECTQueryExecutor(
            code_to_location_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        code_to_location_df = select_statement_executor.execute_query()

        return code_to_location_df

    def clean_resp(self, data):
        clean_data = []
        for k, v in data.items():
            clean_data.extend(v)
        return clean_data

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "postal_code",
            "country_code",
            "latitude",
            "longitude",
            "city",
            "state",
            "city_en",
            "state_en",
            "state_code",
            "province",
            "province_code"
        ]


class ZipCodeBaseCodeInRadiusTable(APITable):
    """The ZipCodeBase Codes within Radius Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://app.zipcodebase.com/documentation#radius API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            codes within the radius

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'codes_within_radius',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []

        for op, arg1, arg2 in where_conditions:
            if arg1 == "code":
                if op == '=':
                    search_params["code"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for code column.")

            if arg1 == "radius":
                if op == '=':
                    search_params["radius"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for radius column.")

            if arg1 == "country":
                if op == '=':
                    search_params["country"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for country column.")

            if arg1 == "unit":
                if op == '=':
                    search_params["unit"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for unit column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("code" in search_params) and ("radius" in search_params) and ("country" in search_params)

        if not filter_flag:
            raise NotImplementedError("`codes`, `radius` and `country` columns have to be present in where clause.")

        code_to_location_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.client.codes_within_radius(search_params.get("code"), search_params.get("radius"), search_params.get("country"), search_params.get("unit", "km"))

        self.check_res(res=response)

        content = response["content"]

        logger.info(f"response size - {len(content['results'])}")
        code_to_location_df = pd.json_normalize(content["results"])

        select_statement_executor = SELECTQueryExecutor(
            code_to_location_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        code_to_location_df = select_statement_executor.execute_query()

        return code_to_location_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "code",
            "city",
            "state",
            "city_en",
            "state_en",
            "distance"
        ]


class ZipCodeBaseCodeByCityTable(APITable):
    """The ZipCodeBase Codes within a City Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://app.zipcodebase.com/documentation#city API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            codes within the city

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'codes_by_city',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []

        for op, arg1, arg2 in where_conditions:
            if arg1 == "city":
                if op == '=':
                    search_params["city"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for city column.")

            if arg1 == "country":
                if op == '=':
                    search_params["country"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for country column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("city" in search_params) and ("country" in search_params)

        if not filter_flag:
            raise NotImplementedError("`city` and `country` columns have to be present in where clause.")

        codes_by_city_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.client.codes_by_city(search_params.get("city"), search_params.get("country"))

        self.check_res(res=response)

        content = response["content"]

        logger.info(f"response size - {len(content['results'])}")
        codes_by_city_df = pd.json_normalize({"codes": content["results"]})

        select_statement_executor = SELECTQueryExecutor(
            codes_by_city_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        codes_by_city_df = select_statement_executor.execute_query()

        return codes_by_city_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "codes"
        ]


class ZipCodeBaseCodeByStateTable(APITable):
    """The ZipCodeBase Codes within a State Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://app.zipcodebase.com/documentation#state API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            codes within the state

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'codes_by_state',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []

        for op, arg1, arg2 in where_conditions:
            if arg1 == "state":
                if op == '=':
                    search_params["state"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for state column.")

            if arg1 == "country":
                if op == '=':
                    search_params["country"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for country column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("state" in search_params) and ("country" in search_params)

        if not filter_flag:
            raise NotImplementedError("`state` and `country` columns have to be present in where clause.")

        codes_by_state_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.client.codes_by_state(search_params.get("state"), search_params.get("country"))

        self.check_res(res=response)

        content = response["content"]

        logger.info(f"response size - {len(content['results'])}")
        codes_by_state_df = pd.json_normalize({"codes": content["results"]})

        select_statement_executor = SELECTQueryExecutor(
            codes_by_state_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        codes_by_state_df = select_statement_executor.execute_query()

        return codes_by_state_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "codes"
        ]


class ZipCodeBaseStatesByCountryTable(APITable):
    """The ZipCodeBase Provinces/states within a country Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://app.zipcodebase.com/documentation#provinces API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            states within a country

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'states_by_country',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []

        for op, arg1, arg2 in where_conditions:

            if arg1 == "country":
                if op == '=':
                    search_params["country"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for country column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("country" in search_params)

        if not filter_flag:
            raise NotImplementedError("`country` column has to be present in where clause.")

        states_by_country_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.client.states_by_country(search_params.get("country"))

        self.check_res(res=response)

        content = response["content"]

        logger.info(f"response size - {len(content['results'])}")
        states_by_country_df = pd.json_normalize({"states": content["results"]})

        select_statement_executor = SELECTQueryExecutor(
            states_by_country_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        states_by_country_df = select_statement_executor.execute_query()

        return states_by_country_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "states"
        ]
